# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.modules
r"""
Incidence algebras
"""
# ****************************************************************************
#       Copyright (C) 2014 Travis Scrimshaw <tscrim at ucdavis.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from copy import copy
from typing import Any

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.categories.algebras import Algebras
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets
from sage.combinat.free_module import CombinatorialFreeModule
from sage.matrix.matrix_space import MatrixSpace


class IncidenceAlgebra(CombinatorialFreeModule):
    r"""
    The incidence algebra of a poset.

    Let `P` be a poset and `R` be a commutative unital associative ring.
    The *incidence algebra* `I_P` is the algebra of functions
    `\alpha \colon P \times P \to R` such that `\alpha(x, y) = 0`
    if `x \not\leq y` where multiplication is given by convolution:

    .. MATH::

        (\alpha \ast \beta)(x, y) = \sum_{x \leq k \leq y}
        \alpha(x, k) \beta(k, y).

    This has a natural basis given by indicator functions for the
    interval `[a, b]`, i.e. `X_{a,b}(x,y) = \delta_{ax} \delta_{by}`.
    The incidence algebra is a unital algebra with the identity given
    by the Kronecker delta `\delta(x, y) = \delta_{xy}`. The Möbius
    function of `P` is another element of `I_p` whose inverse is the
    `\zeta` function of the poset (so `\zeta(x, y) = 1` for
    every interval `[x, y]`).

    .. TODO::

        Implement the incidence coalgebra.

    REFERENCES:

    - :wikipedia:`Incidence_algebra`
    """
    def __init__(self, R, P, prefix='I') -> None:
        """
        Initialize ``self``.

        TESTS::

            sage: P = posets.BooleanLattice(3)
            sage: I = P.incidence_algebra(QQ)
            sage: TestSuite(I).run()  # long time
        """
        cat = Algebras(R).WithBasis()
        if P in FiniteEnumeratedSets():
            cat = cat.FiniteDimensional()
        self._poset = P
        CombinatorialFreeModule.__init__(self, R, map(tuple, P.relations()),
                                         prefix=prefix, category=cat)

    def _repr_term(self, A):
        """
        Return a string representation of the term labeled by ``A``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I._repr_term((4, 12))
            'I[4, 12]'
        """
        return self.prefix() + str(list(A))

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: P.incidence_algebra(QQ)
            Incidence algebra of Finite lattice containing 16 elements
             over Rational Field
        """
        return f"Incidence algebra of {self._poset} over {self.base_ring()}"

    def _coerce_map_from_(self, R):
        """
        Return the coerce map from ``R`` into ``self`` if it exists
        or ``False`` otherwise.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: R = I.reduced_subalgebra()
            sage: I.has_coerce_map_from(R)
            True
            sage: I.has_coerce_map_from(QQ)
            True
            sage: Pp = posets.BooleanLattice(3)
            sage: Rp = Pp.incidence_algebra(QQ).reduced_subalgebra()
            sage: I.has_coerce_map_from(Rp)
            False
        """
        if isinstance(R, ReducedIncidenceAlgebra) and R._ambient is self:
            return copy(R.lift)
        return super()._coerce_map_from_(R)

    def reduced_subalgebra(self, prefix='R'):
        """
        Return the reduced incidence subalgebra.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I.reduced_subalgebra()
            Reduced incidence algebra of Finite lattice containing 16 elements
             over Rational Field
        """
        return ReducedIncidenceAlgebra(self, prefix)

    def poset(self):
        """
        Return the defining poset of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I.poset()
            Finite lattice containing 16 elements
            sage: I.poset() == P
            True
        """
        return self._poset

    def some_elements(self):
        """
        Return a list of elements of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(1)
            sage: I = P.incidence_algebra(QQ)
            sage: Ielts = I.some_elements(); Ielts # random
            [2*I[0, 0] + 2*I[0, 1] + 3*I[1, 1],
             I[0, 0] - I[0, 1] + I[1, 1],
             I[0, 0] + I[0, 1] + I[1, 1]]
            sage: [a in I for a in Ielts]
            [True, True, True]
        """
        return [self.an_element(), self.moebius(), self.zeta()]

    def product_on_basis(self, A, B):
        r"""
        Return the product of basis elements indexed by ``A`` and ``B``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I.product_on_basis((1, 3), (3, 11))
            I[1, 11]
            sage: I.product_on_basis((1, 3), (2, 2))
            0
        """
        if A[1] == B[0]:
            return self.monomial((A[0], B[1]))
        return self.zero()

    @cached_method
    def one(self):
        r"""
        Return the element `1` in ``self`` (which is the Kronecker
        delta `\delta(x, y)`).

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I.one()
            I[0, 0] + I[1, 1] + I[2, 2] + I[3, 3] + I[4, 4] + I[5, 5]
             + I[6, 6] + I[7, 7] + I[8, 8] + I[9, 9] + I[10, 10]
             + I[11, 11] + I[12, 12] + I[13, 13] + I[14, 14] + I[15, 15]
        """
        return self.sum_of_monomials((x, x) for x in self._poset)

    delta = one

    @cached_method
    def zeta(self):
        r"""
        Return the `\zeta` function in ``self``.

        The `\zeta` function on a poset `P` is given by

        .. MATH::

            \zeta(x, y) = \begin{cases}
            1 & x \leq y, \\
            0 & x \not\leq y.
            \end{cases}

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I.zeta() * I.moebius() == I.one()
            True
        """
        return self.sum(self.basis())

    @cached_method
    def moebius(self):
        """
        Return the Möbius function of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(2)
            sage: I = P.incidence_algebra(QQ)
            sage: I.moebius()
            I[0, 0] - I[0, 1] - I[0, 2] + I[0, 3] + I[1, 1]
             - I[1, 3] + I[2, 2] - I[2, 3] + I[3, 3]
        """
        mu = self._poset.moebius_function
        R = self.base_ring()
        return self.sum_of_terms((A, R(mu(*A))) for A in self.basis().keys())

    def __getitem__(self, A):
        """
        Return the basis element indexed by ``A``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I[2, 6]
            I[2, 6]
            sage: I[(2, 6)]
            I[2, 6]
            sage: I[[2, 6]]
            I[2, 6]
            sage: I[2]
            I[2, 2]

            sage: I[2, 5]
            Traceback (most recent call last):
            ...
            ValueError: not an interval

            sage: I[-1]
            Traceback (most recent call last):
            ...
            ValueError: not an element of the poset
        """
        if not isinstance(A, (list, tuple)):
            if A not in self._poset.list():
                raise ValueError("not an element of the poset")
            A = (A, A)
        else:
            A = tuple(A)
            if len(A) != 2 or A not in self.basis().keys():
                raise ValueError("not an interval")
        return self.monomial(A)

    @lazy_attribute
    def _linear_extension(self):
        """
        Return a fixed linear extension of the defining poset of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: I._linear_extension
            (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15)
        """
        return tuple(self._poset.linear_extension())

    class Element(CombinatorialFreeModule.Element):
        """
        An element of an incidence algebra.
        """
        def __call__(self, x, y):
            """
            Return ``self(x, y)``.

            EXAMPLES::

                sage: P = posets.BooleanLattice(4)
                sage: I = P.incidence_algebra(QQ)
                sage: mu = I.moebius()
                sage: mu(0, 12)
                1
                sage: mu(0, 7)
                -1
                sage: mu(15, 0)
                0
            """
            P = self.parent()._poset
            x = P(x)
            y = P(y)
            return self[x, y]

        @cached_method
        def to_matrix(self):
            r"""
            Return ``self`` as a matrix.

            We define a matrix `M_{xy} = \alpha(x, y)` for some element
            `\alpha \in I_P` in the incidence algebra `I_P` and we order
            the elements `x,y \in P` by some linear extension of `P`. This
            defines an algebra (iso)morphism; in particular, multiplication
            in the incidence algebra goes to matrix multiplication.

            EXAMPLES::

                sage: P = posets.BooleanLattice(2)
                sage: I = P.incidence_algebra(QQ)
                sage: I.moebius().to_matrix()
                [ 1 -1 -1  1]
                [ 0  1  0 -1]
                [ 0  0  1 -1]
                [ 0  0  0  1]
                sage: I.zeta().to_matrix()
                [1 1 1 1]
                [0 1 0 1]
                [0 0 1 1]
                [0 0 0 1]

            TESTS:

            We check that this is an algebra (iso)morphism::

                sage: P = posets.BooleanLattice(4)
                sage: I = P.incidence_algebra(QQ)
                sage: mu = I.moebius()
                sage: (mu*mu).to_matrix() == mu.to_matrix() * mu.to_matrix()
                True
            """
            P = self.parent()
            MS = MatrixSpace(P.base_ring(), P._poset.cardinality(), sparse=True)
            L = P._linear_extension
            M = copy(MS.zero())
            for i, c in self:
                M[L.index(i[0]), L.index(i[1])] = c
            M.set_immutable()
            return M

        def is_unit(self):
            """
            Return if ``self`` is a unit.

            EXAMPLES::

                sage: P = posets.BooleanLattice(2)
                sage: I = P.incidence_algebra(QQ)
                sage: mu = I.moebius()
                sage: mu.is_unit()
                True
                sage: zeta = I.zeta()
                sage: zeta.is_unit()
                True
                sage: x = mu - I.zeta() + I[2,2]
                sage: x.is_unit()
                False
                sage: y = I.moebius() + I.zeta()
                sage: y.is_unit()
                True

            This depends on the base ring::

                sage: I = P.incidence_algebra(ZZ)
                sage: y = I.moebius() + I.zeta()
                sage: y.is_unit()
                False
            """
            return all(self[x, x].is_unit() for x in self.parent()._poset)

        def __invert__(self):
            """
            Return the inverse of ``self`` if it exists.

            EXAMPLES::

                sage: P = posets.BooleanLattice(2)
                sage: I = P.incidence_algebra(QQ)
                sage: mu = I.moebius()
                sage: ~mu
                I[0, 0] + I[0, 1] + I[0, 2] + I[0, 3] + I[1, 1]
                 + I[1, 3] + I[2, 2] + I[2, 3] + I[3, 3]
                sage: x = mu - I.zeta() + I[2,2]
                sage: ~x
                Traceback (most recent call last):
                ...
                ValueError: element is not invertible

            TESTS::

                sage: P = posets.BooleanLattice(4)
                sage: I = P.incidence_algebra(QQ)
                sage: mu = I.moebius()
                sage: ~mu == I.zeta()
                True
                sage: ~I.one() == I.one()
                True
            """
            M = self.to_matrix()
            if not M.is_invertible():
                raise ValueError("element is not invertible")
            inv = ~M
            L = self.parent()._linear_extension
            return self.parent().sum_of_terms(
                ((L[i], L[j]), inv[i, j])
                for i, j in inv.nonzero_positions(copy=False))


class ReducedIncidenceAlgebra(CombinatorialFreeModule):
    r"""
    The reduced incidence algebra of a poset.

    The reduced incidence algebra `R_P` is a subalgebra of the
    incidence algebra `I_P` where `\alpha(x, y) = \alpha(x', y')` when
    `[x, y]` is isomorphic to `[x', y']` as posets. Thus the delta, Möbius,
    and zeta functions are all elements of `R_P`.
    """
    def __init__(self, I, prefix='R') -> None:
        """
        Initialize ``self``.

        TESTS::

            sage: P = posets.BooleanLattice(3)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: TestSuite(R).run()  # long time
        """
        self._ambient = I
        EC: dict[Any, list] = {}
        P = self._ambient._poset
        if not P.is_finite():
            raise NotImplementedError("only implemented for finite posets")
        for i in self._ambient.basis().keys():
            S = P.subposet(P.interval(*i))
            added = False
            for k, ECk in EC.items():
                if S._hasse_diagram.is_isomorphic(k._hasse_diagram):
                    ECk.append(i)
                    added = True
                    break
            if not added:
                EC[S] = [i]
        equiv_classes = map(sorted, EC.values())
        self._equiv_classes = {cls[0]: cls for cls in equiv_classes}
        cat = Algebras(I.base_ring()).FiniteDimensional().WithBasis()
        CombinatorialFreeModule.__init__(self, I.base_ring(),
                                         sorted(self._equiv_classes.keys()),
                                         prefix=prefix, category=cat)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: P.incidence_algebra(QQ).reduced_subalgebra()
            Reduced incidence algebra of Finite lattice containing 16 elements
             over Rational Field
        """
        msg = "Reduced incidence algebra of {} over {}"
        return msg.format(self._ambient._poset, self.base_ring())

    def poset(self):
        """
        Return the defining poset of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.poset()
            Finite lattice containing 16 elements
            sage: R.poset() == P
            True
        """
        return self._ambient._poset

    def some_elements(self):
        """
        Return a list of elements of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.some_elements()
            [2*R[(0, 0)] + 2*R[(0, 1)] + 3*R[(0, 3)],
             R[(0, 0)] - R[(0, 1)] + R[(0, 3)] - R[(0, 7)] + R[(0, 15)],
             R[(0, 0)] + R[(0, 1)] + R[(0, 3)] + R[(0, 7)] + R[(0, 15)]]
        """
        return [self.an_element(), self.moebius(), self.zeta()]

    @cached_method
    def one_basis(self):
        """
        Return the index of the element `1` in ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.one_basis()
            (0, 0)
        """
        for A in self.basis().keys():
            if A[0] == A[1]:
                return A

    def delta(self):
        """
        Return the Kronecker delta function in ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.delta()
            R[(0, 0)]
        """
        return self.one()

    @cached_method
    def zeta(self):
        r"""
        Return the `\zeta` function in ``self``.

        The `\zeta` function on a poset `P` is given by

        .. MATH::

            \zeta(x, y) = \begin{cases}
            1 & x \leq y, \\
            0 & x \not\leq y.
            \end{cases}

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.zeta()
            R[(0, 0)] + R[(0, 1)] + R[(0, 3)] + R[(0, 7)] + R[(0, 15)]
        """
        return self.sum(self.basis())

    @cached_method
    def moebius(self):
        """
        Return the Möbius function of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.moebius()
            R[(0, 0)] - R[(0, 1)] + R[(0, 3)] - R[(0, 7)] + R[(0, 15)]
        """
        mu = self._ambient._poset.moebius_function
        R = self.base_ring()
        return self.sum_of_terms((A, R(mu(*A))) for A in self.basis().keys())

    @cached_method
    def _lift_basis(self, x):
        """
        Lift the basis element indexed by ``x`` to the ambient incidence
        algebra of ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R._lift_basis((0, 7))
            I[0, 7] + I[0, 11] + I[0, 13] + I[0, 14] + I[1, 15]
             + I[2, 15] + I[4, 15] + I[8, 15]
        """
        return self._ambient.sum_of_monomials(self._equiv_classes[x])

    @lazy_attribute
    def lift(self):
        """
        Return the lift morphism from ``self`` to the ambient space.

        EXAMPLES::

            sage: P = posets.BooleanLattice(2)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R.lift
            Generic morphism:
              From: Reduced incidence algebra of Finite lattice containing 4 elements over Rational Field
              To:   Incidence algebra of Finite lattice containing 4 elements over Rational Field
            sage: R.an_element() - R.one()
            R[(0, 0)] + 2*R[(0, 1)] + 3*R[(0, 3)]
            sage: R.lift(R.an_element() - R.one())
            I[0, 0] + 2*I[0, 1] + 2*I[0, 2] + 3*I[0, 3] + I[1, 1]
             + 2*I[1, 3] + I[2, 2] + 2*I[2, 3] + I[3, 3]
        """
        return self.module_morphism(self._lift_basis, codomain=self._ambient)

    def __getitem__(self, A):
        """
        Return the basis element indexed by ``A``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
            sage: R[0, 0]
            R[(0, 0)]
            sage: R[0, 1]
            R[(0, 1)]
            sage: R[0, 15]
            R[(0, 15)]
            sage: R[0]
            R[(0, 0)]

        This works for any representative of the equivalence class::

            sage: R[3, 3]
            R[(0, 0)]
            sage: R[3, 11]
            R[(0, 1)]

        TESTS:

            sage: R[2, 5]
            Traceback (most recent call last):
            ...
            ValueError: not an interval

            sage: R[-1]
            Traceback (most recent call last):
            ...
            ValueError: not an element of the poset
        """
        if not isinstance(A, (list, tuple)):
            if A not in self._ambient._poset.list():
                raise ValueError("not an element of the poset")
            return self.one()
        else:
            A = tuple(A)
            if len(A) != 2:
                raise ValueError("not an interval")
        for k in self._equiv_classes:
            if A in self._equiv_classes[k]:
                return self.monomial(k)
        raise ValueError("not an interval")

    def _retract(self, x):
        """
        Return the retract of ``x`` from the incidence algebra to ``self``.

        EXAMPLES::

            sage: P = posets.BooleanLattice(4)
            sage: I = P.incidence_algebra(QQ)
            sage: R = I.reduced_subalgebra()
            sage: all(R._retract(R.lift(x)) == x for x in R.basis())
            True
            sage: R._retract(I.zeta()) == R.zeta()
            True
            sage: R._retract(I.delta()) == R.delta()
            True
            sage: R._retract(I.moebius()) == R.moebius()
            True
        """
        return self.sum_of_terms((k, x[k]) for k in self.basis().keys())

    class Element(CombinatorialFreeModule.Element):
        """
        An element of a reduced incidence algebra.
        """
        def __call__(self, x, y):
            """
            Return ``self(x, y)``.

            EXAMPLES::

                sage: P = posets.BooleanLattice(4)
                sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
                sage: x = R.an_element()
                sage: x(0, 7)
                0
                sage: x(7, 15)
                2
                sage: x(4, 15)
                0
            """
            return self.parent().lift(self)(x, y)

        def _mul_(self, other):
            """
            Return the product of ``self`` and ``other``.

            EXAMPLES::

                sage: P = posets.BooleanLattice(4)
                sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
                sage: x = R.an_element()
                sage: x * R.zeta()
                2*R[(0, 0)] + 4*R[(0, 1)] + 9*R[(0, 3)] + 17*R[(0, 7)] + 28*R[(0, 15)]
                sage: x * R.moebius()
                2*R[(0, 0)] + R[(0, 3)] - 5*R[(0, 7)] + 12*R[(0, 15)]
                sage: x * R.moebius() * R.zeta() == x
                True
            """
            P = self.parent()
            return P._retract(P.lift(self) * P.lift(other))

        @cached_method
        def to_matrix(self):
            r"""
            Return ``self`` as a matrix.

            EXAMPLES::

                sage: P = posets.BooleanLattice(2)
                sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
                sage: mu = R.moebius()
                sage: mu.to_matrix()
                [ 1 -1 -1  1]
                [ 0  1  0 -1]
                [ 0  0  1 -1]
                [ 0  0  0  1]
            """
            return self.parent().lift(self).to_matrix()

        def is_unit(self):
            """
            Return if ``self`` is a unit.

            EXAMPLES::

                sage: P = posets.BooleanLattice(4)
                sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
                sage: x = R.an_element()
                sage: x.is_unit()
                True
            """
            return self[self.parent().one_basis()].is_unit()

        def __invert__(self):
            """
            Return the inverse of ``self``.

            EXAMPLES::

                sage: P = posets.BooleanLattice(4)
                sage: R = P.incidence_algebra(QQ).reduced_subalgebra()
                sage: x = R.an_element()
                sage: ~x
                1/2*R[(0, 0)] - 1/2*R[(0, 1)] + 1/4*R[(0, 3)]
                 + 3/2*R[(0, 7)] - 33/4*R[(0, 15)]
            """
            P = self.parent()
            return P._retract(~P.lift(self))

        def lift(self):
            """
            Return the lift of ``self`` to the ambient space.

            EXAMPLES::

                sage: P = posets.BooleanLattice(2)
                sage: I = P.incidence_algebra(QQ)
                sage: R = I.reduced_subalgebra()
                sage: x = R.an_element(); x
                2*R[(0, 0)] + 2*R[(0, 1)] + 3*R[(0, 3)]
                sage: x.lift()
                2*I[0, 0] + 2*I[0, 1] + 2*I[0, 2] + 3*I[0, 3] + 2*I[1, 1]
                 + 2*I[1, 3] + 2*I[2, 2] + 2*I[2, 3] + 2*I[3, 3]
            """
            return self.parent().lift(self)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs
r"""
Homsets between simplicial complexes

AUTHORS:

- Travis Scrimshaw (2012-08-18): Made all simplicial complexes immutable to
  work with the homset cache.

EXAMPLES::

    sage: S = simplicial_complexes.Sphere(1)
    sage: T = simplicial_complexes.Sphere(2)
    sage: H = Hom(S,T)
    sage: f = {0:0,1:1,2:3}
    sage: x = H(f)
    sage: x
    Simplicial complex morphism:
      From: Minimal triangulation of the 1-sphere
      To: Minimal triangulation of the 2-sphere
      Defn: 0 |--> 0
            1 |--> 1
            2 |--> 3
    sage: x.is_injective()
    True
    sage: x.is_surjective()
    False
    sage: x.image()
    Simplicial complex with vertex set (0, 1, 3) and facets {(0, 1), (0, 3), (1, 3)}
    sage: from sage.topology.simplicial_complex import Simplex
    sage: s = Simplex([1,2])
    sage: x(s)
    (1, 3)

TESTS::

    sage: S = simplicial_complexes.Sphere(1)
    sage: T = simplicial_complexes.Sphere(2)
    sage: H = Hom(S,T)
    sage: loads(dumps(H)) == H
    True
"""

# ****************************************************************************
# Copyright (C) 2009 D. Benjamin Antieau <d.ben.antieau@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
#
# ****************************************************************************

import sage.categories.homset
from .simplicial_complex_morphism import SimplicialComplexMorphism


def is_SimplicialComplexHomset(x) -> bool:
    """
    Return ``True`` if and only if ``x`` is a simplicial complex homspace.

    EXAMPLES::

        sage: S = SimplicialComplex(is_mutable=False)
        sage: T = SimplicialComplex(is_mutable=False)
        sage: H = Hom(S, T)
        sage: H
        Set of Morphisms from Simplicial complex with vertex set () and facets {()}
         to Simplicial complex with vertex set () and facets {()}
         in Category of finite simplicial complexes
        sage: from sage.topology.simplicial_complex_homset import is_SimplicialComplexHomset
        sage: is_SimplicialComplexHomset(H)
        doctest:warning...
        DeprecationWarning: the function is_SimplicialComplexHomset is deprecated;
        use 'isinstance(..., SimplicialComplexHomset)' instead
        See https://github.com/sagemath/sage/issues/37922 for details.
        True
    """
    from sage.misc.superseded import deprecation
    deprecation(37922, "the function is_SimplicialComplexHomset is deprecated; use 'isinstance(..., SimplicialComplexHomset)' instead")
    return isinstance(x, SimplicialComplexHomset)


class SimplicialComplexHomset(sage.categories.homset.Homset):
    def __call__(self, f):
        """
        INPUT:

        - ``f`` -- dictionary with keys exactly the vertices of the domain
          and values vertices of the codomain

        EXAMPLES::

            sage: S = simplicial_complexes.Sphere(3)
            sage: T = simplicial_complexes.Sphere(2)
            sage: f = {0:0,1:1,2:2,3:2,4:2}
            sage: H = Hom(S,T)
            sage: x = H(f)
            sage: x
            Simplicial complex morphism:
              From: Minimal triangulation of the 3-sphere
              To: Minimal triangulation of the 2-sphere
              Defn: [0, 1, 2, 3, 4] --> [0, 1, 2, 2, 2]
        """
        return SimplicialComplexMorphism(f, self.domain(), self.codomain())

    def diagonal_morphism(self, rename_vertices=True):
        r"""
        Return the diagonal morphism in `Hom(S, S \times S)`.

        EXAMPLES::

            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S,S.product(S, is_mutable=False))
            sage: d = H.diagonal_morphism(); d
            Simplicial complex morphism:
              From: Minimal triangulation of the 2-sphere
              To:   Simplicial complex with 16 vertices and 96 facets
              Defn: 0 |--> L0R0
                    1 |--> L1R1
                    2 |--> L2R2
                    3 |--> L3R3

            sage: T = SimplicialComplex([[0], [1]], is_mutable=False)
            sage: U = T.product(T, rename_vertices=False, is_mutable=False)
            sage: G = Hom(T, U)
            sage: e = G.diagonal_morphism(rename_vertices=False); e
            Simplicial complex morphism:
              From: Simplicial complex with vertex set (0, 1) and facets {(0,), (1,)}
              To:   Simplicial complex with 4 vertices and
                    facets {((0, 0),), ((0, 1),), ((1, 0),), ((1, 1),)}
              Defn: 0 |--> (0, 0)
                    1 |--> (1, 1)
        """
        # Preserve whether the codomain is mutable when renaming the vertices.
        mutable = self._codomain.is_mutable()
        X = self._domain.product(self._domain, rename_vertices=rename_vertices,
                                 is_mutable=mutable)
        if self._codomain != X:
            raise TypeError("diagonal morphism is only defined for Hom(X,XxX)")
        f = {}
        if rename_vertices:
            f = {i: "L{0}R{0}".format(i) for i in self._domain.vertices()}
        else:
            f = {i: (i, i) for i in self._domain.vertices()}
        return SimplicialComplexMorphism(f, self._domain, X)

    def identity(self):
        """
        Return the identity morphism of `Hom(S,S)`.

        EXAMPLES::

            sage: S = simplicial_complexes.Sphere(2)
            sage: H = Hom(S, S)
            sage: i = H.identity()
            sage: i.is_identity()
            True

            sage: T = SimplicialComplex([[0,1]], is_mutable=False)
            sage: G = Hom(T, T)
            sage: G.identity()
            Simplicial complex endomorphism of
             Simplicial complex with vertex set (0, 1) and facets {(0, 1)}
              Defn: 0 |--> 0
                    1 |--> 1
        """
        if not self.is_endomorphism_set():
            raise TypeError("identity map is only defined for endomorphism sets")
        f = {i: i for i in self._domain.vertices()}
        return SimplicialComplexMorphism(f, self._domain, self._codomain)

    def an_element(self):
        """
        Return a (non-random) element of ``self``.

        EXAMPLES::

            sage: S = simplicial_complexes.KleinBottle()
            sage: T = simplicial_complexes.Sphere(5)
            sage: H = Hom(S,T)
            sage: x = H.an_element()
            sage: x
            Simplicial complex morphism:
              From: Minimal triangulation of the Klein bottle
              To:   Minimal triangulation of the 5-sphere
              Defn: [0, 1, 2, 3, 4, 5, 6, 7] --> [0, 0, 0, 0, 0, 0, 0, 0]
        """
        X_vertices = self._domain.vertices()
        try:
            i = next(iter(self._codomain.vertices()))
        except StopIteration:
            if not X_vertices:
                return {}
            else:
                raise TypeError("there are no morphisms from a non-empty simplicial complex to an empty simplicial complex")
        f = {x: i for x in X_vertices}
        return SimplicialComplexMorphism(f, self._domain, self._codomain)



# sage_setup: distribution = sagemath-graphs
# cython: binding=True
r"""
Static dense graphs

This module gathers everything which is related to static dense graphs, i.e. :

- The vertices are integer from `0` to `n-1`
- No labels on vertices/edges
- No multiple edges
- No addition/removal of vertices

This being said, it is technically possible to add/remove edges. The data
structure does not mind at all.

It is all based on the binary matrix data structure described in
``data_structures/binary_matrix.pxd``, which is almost a copy of the bitset data
structure. The only difference is that it differentiates the rows (the vertices)
instead of storing the whole data in a long bitset, and we can use that.

For an overview of graph data structures in sage, see
:mod:`~sage.graphs.base.overview`.

Index
-----

**Cython functions**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    ``dense_graph_init`` | Fill a binary matrix with the information from a Sage (di)graph.

**Python functions**

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`is_strongly_regular` | Check whether the graph is strongly regular
    :meth:`is_triangle_free` | Check whether `G` is triangle free
    :meth:`triangles_count` | Return the number of triangles containing `v`, for every `v`
    :meth:`connected_subgraph_iterator` | Iterator over the induced connected subgraphs of order at most `k`

Functions
---------
"""
from sage.data_structures.binary_matrix cimport *
from cysignals.signals cimport sig_on, sig_off, sig_check
from memory_allocator cimport MemoryAllocator
from itertools import product
from sage.misc.flatten import flatten


cdef dict dense_graph_init(binary_matrix_t m, g, translation=None, force_undirected=False):
    r"""
    Fill a binary matrix with the information from a Sage (di)graph.

    INPUT:

    - ``binary_matrix_t m`` -- the binary matrix to be filled

    - ``g`` -- a graph or digraph

    - ``translation`` -- (default: ``None``) several options for this parameter
      used to specify the mapping from vertices to integers:

      - ``True``, ``False``, ``None`` -- the `i`-th vertex in the binary matrix
        corresponds to vertex ``g.vertices(sort=True)[i]``.
        When set to ``True``, a dictionary encoding the mapping from the
        vertices of `g` to integers in `(0, \dots, n-1)` is returned.

      - a ``list`` -- defines a mapping from integers in `(0, \dots, n-1)` to
        the vertices in the graph `g`. So the `i`-th vertex in the binary matrix
        corresponds to vertex ``translation[i]``.
        **Beware that no checks are made that this input is correct**.

      - a ``dict`` -- defines a mapping from the vertices of the graph to
        integers in `(0, \dots, n-1)`. So the `i`-th vertex in the binary matrix
        corresponds to ``translation[v]`` for some vertex `v \in g`.
        **Beware that no checks are made that this input is correct**.

    - ``force_undirected`` -- boolean (default: ``False``); whether to consider
      the graph as undirected or not
    """
    cdef dict d_translation = {}
    from sage.graphs.graph import Graph
    cdef bint is_undirected = isinstance(g, Graph) or force_undirected
    cdef int n = g.order()
    cdef int i, j

    binary_matrix_init(m, n, n)

    if translation and translation is not True:
        if isinstance(translation, list):
            # We are given a mapping from integers to vertices
            d_translation = {v: i for i, v in enumerate(translation)}
        elif isinstance(translation, dict):
            # We are given a mapping vertices to integers
            d_translation = translation

    # If the vertices are 0...n-1, let's avoid an unnecessary dictionary
    if not d_translation and set(g.vertex_iterator()) == set(range(n)):
        if translation is True:
            d_translation = {i: i for i in range(n)}

        for i, j in g.edge_iterator(labels=False):
            binary_matrix_set1(m, i, j)
            if is_undirected:
                binary_matrix_set1(m, j, i)
    else:
        if not d_translation:
            d_translation = {v: i for i, v in enumerate(g.vertices(sort=True))}

        for u, v in g.edge_iterator(labels=False):
            binary_matrix_set1(m, d_translation[u], d_translation[v])
            if is_undirected:
                binary_matrix_set1(m, d_translation[v], d_translation[u])

    if translation is True:
        return d_translation


def is_strongly_regular(g, parameters=False):
    r"""
    Check whether the graph is strongly regular.

    A simple graph `G` is said to be strongly regular with parameters
    `(n, k, \lambda, \mu)` if and only if:

    * `G` has `n` vertices

    * `G` is `k`-regular

    * Any two adjacent vertices of `G` have `\lambda` common neighbors

    * Any two non-adjacent vertices of `G` have `\mu` common neighbors

    By convention, the complete graphs, the graphs with no edges and the empty
    graph are not strongly regular.

    See the :wikipedia:`Strongly regular graph`.

    INPUT:

    - ``parameters`` -- boolean (default: ``False``); whether to return the
      quadruple `(n, k, \lambda, \mu)`. If ``parameters = False`` (default),
      this method only returns ``True`` and ``False`` answers.
      If ``parameters = True``, the ``True`` answers are replaced by quadruples
      `(n, k, \lambda, \mu)`. See definition above.

    EXAMPLES:

    Petersen's graph is strongly regular::

        sage: g = graphs.PetersenGraph()
        sage: g.is_strongly_regular()
        True
        sage: g.is_strongly_regular(parameters=True)
        (10, 3, 0, 1)

    And Clebsch's graph is too::

        sage: g = graphs.ClebschGraph()
        sage: g.is_strongly_regular()
        True
        sage: g.is_strongly_regular(parameters=True)
        (16, 5, 0, 2)

    But Chvatal's graph is not::

        sage: g = graphs.ChvatalGraph()
        sage: g.is_strongly_regular()
        False

    Complete graphs are not strongly regular. (:issue:`14297`) ::

        sage: g = graphs.CompleteGraph(5)
        sage: g.is_strongly_regular()
        False

    Completements of complete graphs are not strongly regular::

        sage: g = graphs.CompleteGraph(5).complement()
        sage: g.is_strongly_regular()
        False

    The empty graph is not strongly regular::

        sage: g = graphs.EmptyGraph()
        sage: g.is_strongly_regular()
        False

    If the input graph has loops or multiedges an exception is raised::

        sage: Graph([(1,1),(2,2)],loops=True).is_strongly_regular()
        Traceback (most recent call last):
        ...
        ValueError: This method is not known to work on graphs with
        loops. Perhaps this method can be updated to handle them, but in the
        meantime if you want to use it please disallow loops using
        allow_loops().

        sage: Graph([(1,2),(1,2)],multiedges=True).is_strongly_regular()
        Traceback (most recent call last):
        ...
        ValueError: This method is not known to work on graphs with
        multiedges. Perhaps this method can be updated to handle them, but in
        the meantime if you want to use it please disallow multiedges using
        allow_multiple_edges().
    """
    g._scream_if_not_simple()
    cdef binary_matrix_t m
    cdef bitset_t b_tmp
    cdef int n = g.order()
    cdef int inter
    cdef int i, j, k

    if not g.order() or not g.size():  # no vertices or no edges
        return False

    if g.is_clique():
        return False

    cdef list degree = g.degree()
    k = degree[0]
    if any(d != k for d in degree):
        return False

    bitset_init(b_tmp, n)

    # m is now our copy of the graph
    dense_graph_init(m, g, translation={v: i for i, v in enumerate(g)})

    cdef int llambda = -1
    cdef int mu = -1

    for i in range(n):
        for j in range(i + 1, n):

            # The intersection of the common neighbors of i and j is a AND of
            # their respective rows. A popcount then returns its cardinality.
            bitset_and(b_tmp, m.rows[i], m.rows[j])
            inter = bitset_len(b_tmp)

            # Check that this cardinality is correct according to the values of lambda and mu
            if binary_matrix_get(m, i, j):
                if llambda == -1:
                    llambda = inter
                elif llambda != inter:
                    binary_matrix_free(m)
                    bitset_free(b_tmp)
                    return False
            else:
                if mu == -1:
                    mu = inter
                elif mu != inter:
                    binary_matrix_free(m)
                    bitset_free(b_tmp)
                    return False

    binary_matrix_free(m)
    bitset_free(b_tmp)

    if parameters:
        return (n, k, llambda, mu)
    else:
        return True


def is_triangle_free(G, certificate=False):
    r"""
    Check whether `G` is triangle free.

    INPUT:

    - ``G`` -- a Sage graph

    - ``certificate`` -- boolean (default: ``False``); whether to return a
      triangle if one is found

    EXAMPLES::

        sage: from sage.graphs.base.static_dense_graph import is_triangle_free
        sage: is_triangle_free(graphs.PetersenGraph())
        True
        sage: K4 = graphs.CompleteGraph(4)
        sage: is_triangle_free(K4)
        False
        sage: b, certif = is_triangle_free(K4, certificate=True)
        sage: K4.subgraph(certif).is_clique()
        True

    TESTS::

        sage: from sage.graphs.base.static_dense_graph import is_triangle_free
        sage: is_triangle_free(Graph())
        True
        sage: is_triangle_free(Graph(), certificate=True)
        (True, [])
    """
    G._scream_if_not_simple()
    cdef int n = G.order()
    if n < 3:
        return (True, []) if certificate else True

    cdef binary_matrix_t g
    cdef list int_to_vertex = list(G)
    dense_graph_init(g, G, translation=int_to_vertex)

    cdef mp_size_t i, j, k
    for i in range(n):
        j = bitset_next(g.rows[i], i + 1)
        while j != -1:
            if bitset_are_disjoint(g.rows[i], g.rows[j]):
                j = bitset_next(g.rows[i], j + 1)
            else:
                if certificate:
                    # Search for a common neighbor
                    k = bitset_first(g.rows[i])
                    while bitset_not_in(g.rows[j], k):
                        k = bitset_next(g.rows[j], k + 1)
                    certif = [int_to_vertex[i], int_to_vertex[j], int_to_vertex[k]]

                binary_matrix_free(g)
                return (False, certif) if certificate else False

    binary_matrix_free(g)
    return (True, []) if certificate else True


def triangles_count(G):
    r"""
    Return the number of triangles containing `v`, for every `v`.

    INPUT:

    - ``G`` -- a simple Sage graph

    EXAMPLES::

        sage: from sage.graphs.base.static_dense_graph import triangles_count
        sage: triangles_count(graphs.PetersenGraph())
        {0: 0, 1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0}
        sage: sum(triangles_count(graphs.CompleteGraph(15)).values()) == 3 * binomial(15, 3)        # needs sage.symbolic
        True
    """
    from sage.rings.integer import Integer
    G._scream_if_not_simple()
    cdef int n = G.order()

    cdef uint64_t * count = <uint64_t *> check_calloc(n, sizeof(uint64_t))

    cdef binary_matrix_t g
    cdef list int_to_vertex = list(G)
    dense_graph_init(g, G, translation=int_to_vertex)

    cdef bitset_t b_tmp
    bitset_init(b_tmp, n)

    cdef int i, j
    cdef uint64_t tmp_count = 0

    for i in range(n):
        for j in range(i + 1, n):
            if not bitset_in(g.rows[i], j):
                continue
            bitset_and(b_tmp, g.rows[i], g.rows[j])
            tmp_count = bitset_len(b_tmp)
            count[i] += tmp_count
            count[j] += tmp_count

    ans = {v: Integer(count[i] // 2) for i, v in enumerate(int_to_vertex)}

    bitset_free(b_tmp)
    binary_matrix_free(g)
    sig_free(count)

    return ans


def _format_result(G, edges, edges_only, labels):
    r"""
    Helper method for ``connected_full_subgraphs`` to return a result.

    INPUT:

    - ``G`` -- a :class:`DiGraph`

    - ``edges`` -- list of edges ignoring the orientation

    - ``edges_only`` -- boolean; whether to return DiGraph or list of vertices

    - ``labels`` -- boolean; whether to return labeled edges or not. This
      parameter is used only when ``edges_only`` is ``True``.

    EXAMPLES:

    The complete graph of order 3 has 4 connected subgraphs::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.CompleteGraph(3)
        sage: len(list(connected_full_subgraphs(G)))
        4
    """
    if edges_only:
        if labels:
            return [(u, v, G.edge_label(u, v)) for u, v in edges]
        else:
            return edges
    else:
        return G.subgraph(vertices=G, edges=edges)


def _yield_results_for_digraph(G, edges, edges_only, labels, min_edges, max_edges):
    r"""
    Helper method for ``connected_full_subgraphs`` to yield all subdigraphs.

    INPUT:

    - ``G`` -- a :class:`DiGraph`

    - ``edges`` -- list of edges ignoring the orientation

    - ``edges_only`` -- boolean; whether to return DiGraph or list of vertices

    - ``labels`` -- boolean; whether to return labeled edges or not. This
      parameter is used only when ``edges_only`` is ``True``.

    - ``min_edges`` -- integer; minimum number of edges of reported subgraphs

    - ``max_edges`` -- integer; maximum number of edges of reported subgraphs

    EXAMPLES::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = digraphs.Complete(3)
        sage: len(list(connected_full_subgraphs(G)))
        54
    """
    if not edges:
        return
    L = []
    for u, v in edges:
        tmp = []
        if G.has_edge(u, v):
            tmp.append([(u, v)])
        if G.has_edge(v, u):
            tmp.append([(v, u)])
            if G.has_edge(u, v):
                tmp.append([(u, v), (v, u)])
        L.append(tmp)

    if len(L) == 1:
        for F in L[0]:
            if min_edges <= len(F) and len(F) <= max_edges:
                yield _format_result(G, F, edges_only, labels)
    else:
        for E in product(*L):
            F = [e for le in E for e in le]
            if min_edges <= len(F) and len(F) <= max_edges:
                yield _format_result(G, F, edges_only, labels)
    return


def connected_full_subgraphs(G, edges_only=False, labels=False,
                             min_edges=None, max_edges=None):
    r"""
    Return an iterator over the connected subgraphs of `G` with same vertex set.

    This method implements a iterator over the connected subgraphs of the input
    (di)graph with the same ground set of vertices. That is, it iterates over
    every subgraph `H = (V_H, E_H)` of `G = (V, E)` such that `V_H = V`,
    `E_H \subseteq E` and `H` is connected. Hence, this method may yield a huge
    number of graphs.

    When the input (di)graph `G` is not connected, this method returns nothing.

    As for method :meth:`sage.graphs.generic_graph.connected_components`, edge
    orientation is ignored. Hence, the directed graph with a single arc `0 \to
    1` is considered connected.

    INPUT:

    - ``G`` -- a :class:`Graph` or a :class:`DiGraph`; loops and multiple edges
      are *not* allowed

    - ``edges_only`` -- boolean (default: ``False``); whether to return
      (Di)Graph or list of vertices

    - ``labels`` -- boolean (default: ``False``); whether to return labeled
      edges or not. This parameter is used only when ``edges_only`` is ``True``.

    - ``min_edges`` -- integer (default: ``None``); minimum number of edges of
      reported subgraphs. By default (``None``), this lower bound will be set to
      `n - 1`.

    - ``max_edges`` -- integer (default: ``None``); maximum number of edges of
      reported subgraphs. By default (``None``), this lower bound will be set to
      the number of edges of the input (di)graph.

    .. NOTE::

        Roughly, this method explores all possible subsets of neighbors of each
        vertex, which represents a huge number of subsets. We have thus chosen
        to limit the degree of the vertices of the graphs that can be
        considered, even if the graph has a single connected subgraph (e.g., a
        tree). It is therefore recommended to call this method on biconnected
        components, as done in :meth:`connected_subgraph_iterator`.

    EXAMPLES:

    The complete graph of order 3 has 4 connected subgraphs::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.CompleteGraph(3)
        sage: len(list(connected_full_subgraphs(G)))
        4

    A cycle of order 5 has 6 connected subgraphs::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.CycleGraph(5)
        sage: len(list(connected_full_subgraphs(G)))
        6

    The House graph has 18 connected subgraphs of order 5::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.HouseGraph()
        sage: L = list(connected_full_subgraphs(G))
        sage: len(L)
        18
        sage: all(g.order() == 5 for g in L)
        True
        sage: all(g.is_connected() for g in L)
        True
        sage: F = frozenset(frozenset(g.edges(sort=False, labels=False)) for g in L)
        sage: len(F)
        18

    Specifying bounds on the number of edges::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.HouseGraph()
        sage: [g.size() for g in connected_full_subgraphs(G)]
        [6, 5, 5, 5, 4, 4, 5, 4, 4, 4, 5, 4, 4, 4, 5, 4, 4, 4]
        sage: [g.size() for g in connected_full_subgraphs(G, max_edges=4)]
        [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
        sage: [g.size() for g in connected_full_subgraphs(G, min_edges=6)]
        [6]
        sage: [g.size() for g in connected_full_subgraphs(G, min_edges=5, max_edges=5)]
        [5, 5, 5, 5, 5, 5]

    Asking for edges only::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = Graph([(0, 1, "01"), (0, 2, "02"), (1, 2, "12")])
        sage: it = connected_full_subgraphs(G, edges_only=True)
        sage: next(it)
        [(0, 1), (0, 2), (1, 2)]
        sage: next(it)
        [(0, 1), (0, 2)]
        sage: it = connected_full_subgraphs(G, edges_only=True, labels=True)
        sage: next(it)
        [(0, 1, '01'), (0, 2, '02'), (1, 2, '12')]
        sage: next(it)
        [(0, 1, '01'), (0, 2, '02')]

    Subgraphs of a digraph::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = digraphs.Complete(2)
        sage: list(connected_full_subgraphs(G, edges_only=True))
        [[(0, 1)], [(1, 0)], [(0, 1), (1, 0)]]

    TESTS:

    Non connected input graph::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: list(connected_full_subgraphs(Graph(2)))
        Traceback (most recent call last):
        ...
        ValueError: the input (di)graph is not connected

    Too large degree::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.StarGraph(100)
        sage: list(connected_full_subgraphs(G))
        Traceback (most recent call last):
        ...
        ValueError: the degree of the graph is too large for this method

    Wrong bounds on the number of edges::

        sage: from sage.graphs.base.static_dense_graph import connected_full_subgraphs
        sage: G = graphs.HouseGraph()
        sage: [g.size() for g in connected_full_subgraphs(G, min_edges=G.size() + 1)]
        Traceback (most recent call last):
        ...
        ValueError: we must have 4 <= min_edges <= max_edges <= 6
        sage: [g.size() for g in connected_full_subgraphs(G, max_edges=G.order() - 2)]
        Traceback (most recent call last):
        ...
        ValueError: we must have 4 <= min_edges <= max_edges <= 6
    """
    G._scream_if_not_simple()
    if not G.is_connected():
        raise ValueError("the input (di)graph is not connected")

    for d in G.degree():
        if d >= 8 * sizeof(unsigned long) - 1:
            raise ValueError("the degree of the graph is too large for this method")

    cdef Py_ssize_t n = G.order()
    cdef Py_ssize_t m = G.size()
    if min_edges is None or min_edges < n - 1:
        min_edges = n - 1
    if max_edges is None or max_edges > m:
        max_edges = m
    if min_edges > max_edges:
        raise ValueError("we must have {} <= min_edges <= max_edges <= {}".format(n -1, m))

    if n <= 1 or (not G.is_directed() and n == 2) or min_edges == m:
        if edges_only:
            yield list(G.edges(sort=False, labels=labels))
        else:
            yield G.copy()
        return

    # We map vertices to integers. We sort them by degree as a heuristic to
    # reduce the number of operations.
    cdef list int_to_vertex = sorted(G, key=G.degree)
    cdef binary_matrix_t DG
    sig_on()
    dense_graph_init(DG, G, translation=int_to_vertex, force_undirected=True)

    cdef MemoryAllocator mem = MemoryAllocator()
    cdef int * order = <int *> mem.calloc(n, sizeof(int))
    cdef mp_bitcnt_t * n_cpt = <mp_bitcnt_t *> mem.calloc(n, sizeof(mp_bitcnt_t))

    # We use several bitsets to store the current boundary and active neighbors.
    # We also need another bitset that we create at the same time
    cdef binary_matrix_t boundaries
    binary_matrix_init(boundaries, n + 1, n)
    cdef binary_matrix_t neighborhoods
    binary_matrix_init(neighborhoods, n, n)
    sig_off()

    cdef bitset_t active = boundaries.rows[n]  # remaining vertices to consider
    cdef bitset_t boundary  # neighbors of the current subset

    # Initialize the process
    cdef Py_ssize_t i = 0
    order[0] = 0
    bitset_complement(active, active)
    bitset_discard(active, 0)
    bitset_copy(neighborhoods.rows[0], DG.rows[0])
    n_cpt[0] = 1 << bitset_len(DG.rows[0])

    cdef long u, v, j
    cdef mp_bitcnt_t c
    cdef Py_ssize_t num_edges = 0
    cdef list E = []
    cdef list edges

    while i >= 0:
        sig_check()
        if i == n - 1 and num_edges >= min_edges:
            # yield the current solution
            edges = [(int_to_vertex[u], int_to_vertex[v]) for le in E for u, v in le]
            if G.is_directed():
                yield from _yield_results_for_digraph(G, edges, edges_only, labels, min_edges, max_edges)
            else:
                yield _format_result(G, edges, edges_only, labels)

        if n_cpt[i] > 1:
            # Consider the next neighborhood of the current vertex.
            # If vertex u has k active neighbors, we have 2^k possibilities. We
            # use the binary representation of n_cpt[i] to indicate which of the
            # k neighbors are selected. We omit the empty neighborhood which is
            # considered elsewhere.
            n_cpt[i] -= 1
            c = n_cpt[i]
            if num_edges + _bitset_len(&c, 1) > max_edges:
                # Too many edges
                continue

            boundary = boundaries.rows[i + 1]
            bitset_copy(boundary, boundaries.rows[i])
            edges = []
            j = bitset_first(neighborhoods.rows[i])
            while c and j != -1:
                if c & 1:
                    bitset_add(boundary, j)
                    edges.append((order[i], j))
                c >>= 1
                j = bitset_next(neighborhoods.rows[i], j + 1)

            if not bitset_len(boundary):
                # We cannot extend
                continue

            E.append(edges)
            num_edges += len(edges)
            # otherwise, we select a vertex from the boundary and extend the order

        elif bitset_len(boundaries.rows[i]):
            # We prepare the boundary for the selection of the next vertex.
            # This is equivalent to consider an empty neighborhood.
            bitset_copy(boundaries.rows[i + 1], boundaries.rows[i])
            bitset_clear(boundaries.rows[i])  # to prevent doing twice this operation
            E.append([])

        else:
            # We have considered all possible extensions
            bitset_add(active, order[i])
            i -= 1
            if E:
                num_edges -= len(E[-1])
                E.pop()
            continue

        # We select a vertex from the boundary to add to the order
        i += 1
        boundary = boundaries.rows[i]
        u = bitset_first(boundary)
        order[i] = u
        bitset_discard(boundary, u)
        bitset_discard(active, u)
        # prepare neighborhood of u
        bitset_and(neighborhoods.rows[i], active, DG.rows[u])
        j = bitset_len(neighborhoods.rows[i])
        n_cpt[i] = bool(j) << j  # 0 if not j else 2^j

    sig_on()
    binary_matrix_free(boundaries)
    binary_matrix_free(neighborhoods)
    binary_matrix_free(DG)
    sig_off()


def connected_subgraph_iterator(G, k=None, bint vertices_only=False,
                                edges_only=False, labels=False, induced=True,
                                exactly_k=False):
    r"""
    Return an terator over the induced connected subgraphs of order at most `k`.

    This method implements a iterator over the induced connected subgraphs of
    the input (di)graph. An induced subgraph of a graph is another graph, formed
    from a subset of the vertices of the graph and all of the edges connecting
    pairs of vertices in that subset (:wikipedia:`Induced_subgraph`).

    As for method :meth:`sage.graphs.generic_graph.connected_components`, edge
    orientation is ignored. Hence, the directed graph with a single arc `0 \to
    1` is considered connected.

    INPUT:

    - ``G`` -- a :class:`Graph` or a :class:`DiGraph`; loops and multiple edges
      are allowed

    - ``k`` -- (optional) integer; maximum order of the connected subgraphs to
      report; by default, the method iterates over all connected subgraphs
      (equivalent to ``k == n``)

    - ``vertices_only`` -- boolean (default: ``False``); whether to return
      (Di)Graph or list of vertices. This parameter is ignored when ``induced``
      is ``True``.

    - ``edges_only`` -- boolean (default: ``False``); whether to
      return (Di)Graph or list of edges. When ``vertices_only`` is
      ``True``, this parameter is ignored.

    - ``labels`` -- boolean (default: ``False``); whether to return labeled
      edges or not. This parameter is used only when ``vertices_only`` is
      ``False`` and ``edges_only`` is ``True``.

    - ``induced`` -- boolean (default: ``True``); whether to return induced
      connected sub(di)graph only or also non-induced sub(di)graphs.
      This parameter can be set to ``False`` for simple (di)graphs only.

    - ``exactly_k`` -- boolean (default: ``False``); ``True`` if we only
      return graphs of order `k`, ``False`` if we return graphs of order
      at most `k`.

    EXAMPLES::

        sage: G = DiGraph([(1, 2), (2, 3), (3, 4), (4, 2)])
        sage: list(G.connected_subgraph_iterator())
        [Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 3 vertices,
         Subgraph of (): Digraph on 4 vertices,
         Subgraph of (): Digraph on 3 vertices,
         Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 3 vertices,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex]
        sage: list(G.connected_subgraph_iterator(vertices_only=True))
        [[1], [1, 2], [1, 2, 3], [1, 2, 3, 4], [1, 2, 4],
         [2], [2, 3], [2, 3, 4], [2, 4], [3], [3, 4], [4]]
        sage: list(G.connected_subgraph_iterator(k=2))
        [Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex]
        sage: list(G.connected_subgraph_iterator(k=3, vertices_only=True, exactly_k=True))
        [[1, 2, 3], [1, 2, 4], [2, 3, 4]]
        sage: list(G.connected_subgraph_iterator(k=2, vertices_only=True))
        [[1], [1, 2], [2], [2, 3], [2, 4], [3], [3, 4], [4]]

        sage: G = DiGraph([(1, 2), (2, 1)])
        sage: list(G.connected_subgraph_iterator())
        [Subgraph of (): Digraph on 1 vertex,
         Subgraph of (): Digraph on 2 vertices,
         Subgraph of (): Digraph on 1 vertex]
        sage: list(G.connected_subgraph_iterator(vertices_only=True))
        [[1], [1, 2], [2]]

        sage: G = graphs.CompleteGraph(3)
        sage: len(list(G.connected_subgraph_iterator()))
        7
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True)))
        7
        sage: len(list(G.connected_subgraph_iterator(edges_only=True)))
        7
        sage: len(list(G.connected_subgraph_iterator(induced=False)))
        10

        sage: G = DiGraph([(0, 1), (1, 0), (1, 2), (2, 1)])
        sage: len(list(G.connected_subgraph_iterator()))
        6
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True)))
        6
        sage: len(list(G.connected_subgraph_iterator(edges_only=True)))
        6
        sage: len(list(G.connected_subgraph_iterator(induced=False)))
        18

    TESTS:

    The Path Graph of order `n` has `n (n + 1) / 2` connected subgraphs::

        sage: G = graphs.PathGraph(10)
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True)))
        55
        sage: len(list(G.connected_subgraph_iterator(vertices_only=False)))
        55

    The Complete Graph of order `n` has `2^n - 1` connected subgraphs::

        sage: G = graphs.CompleteGraph(5)
        sage: len(list(G.connected_subgraph_iterator(vertices_only=False))) == 2**G.order() - 1
        True
        sage: G = graphs.CompleteGraph(6)
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True))) == 2**G.order() - 1
        True

    Checks that it works with general graphs and corner cases::

        sage: G = DiGraph([(1, 2), (1, 2)], multiedges=True)
        sage: len(list(G.connected_subgraph_iterator()))
        3
        sage: len(list(G.connected_subgraph_iterator(k=0)))
        0
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True)))
        3

        sage: G = Graph([(1, 2), (1, 1)], loops=True)
        sage: len(list(G.connected_subgraph_iterator(vertices_only=False)))
        3
        sage: G = Graph([(1, 2), (1, 2), (1, 1)], loops=True, multiedges=True)
        sage: len(list(G.connected_subgraph_iterator()))
        3
        sage: len(list(G.connected_subgraph_iterator(k=1)))
        2
        sage: len(list(G.connected_subgraph_iterator(vertices_only=True)))
        3
    """
    if not induced:
        G._scream_if_not_simple()
        vertices_only = False

    cdef Py_ssize_t mk = G.order() if k is None else k
    cdef Py_ssize_t n = G.order()
    if not n or mk < 1:
        return

    cdef list int_to_vertex = list(G)
    cdef binary_matrix_t DG
    sig_on()
    dense_graph_init(DG, G, translation=int_to_vertex, force_undirected=True)

    # We use a stack of bitsets. We need 3 bitsets per level with at most n + 1
    # levels, so 3 * n + 3 bitsets. We also need 1 bitset that we create at the
    # same time, so we need overall 3 * n + 4 bitsets
    cdef binary_matrix_t stack
    binary_matrix_init(stack, 3 * n + 4, n)
    sig_off()

    cdef bitset_t current  # current subset of vertices
    cdef bitset_t left     # remaining vertices to consider
    cdef bitset_t boundary  # neighbors of the current subset
    # candidate vertices for extending the current subset, i.e., vertices that
    # are both in left and in boundary
    cdef bitset_t candidates = stack.rows[3 * n + 3]

    cdef Py_ssize_t level
    cdef Py_ssize_t u, v, a
    cdef list vertices

    # We first generate subsets containing vertex 0, then the subsets containing
    # vertex 1 but not vertex 0 since we have already generated all subsets
    # containing 0, then subsets containing 2 but not 0 or 1, etc.
    for u in range(n):
        sig_check()

        vertices = [int_to_vertex[u]]
        if not exactly_k or mk == 1:
            if vertices_only:
                yield vertices
            else:
                H = G.subgraph(vertices)
                if edges_only:
                    yield H.edges(sort=False, labels=labels)
                else:
                    yield H

        # We initialize the loop with vertices u in current, {u+1, ..., n-1}
        # in left, and N(u) in boundary
        bitset_clear(stack.rows[0])
        bitset_add(stack.rows[0], u)
        bitset_set_first_n(stack.rows[1], u + 1)
        bitset_complement(stack.rows[1], stack.rows[1])
        bitset_copy(stack.rows[2], DG.rows[u])
        level = 0

        while level >= 0:
            sig_check()

            # We take the values at the top of the stack
            current = stack.rows[level]
            left = stack.rows[level + 1]
            boundary = stack.rows[level + 2]

            bitset_and(candidates, left, boundary)

            # Search for a candidate vertex v
            v = bitset_next(candidates, u + 1)
            if v >= 0 and bitset_len(current) < mk:
                # We select vertex v
                bitset_discard(left, v)

                # Since we have not modified 'level', the bitsets for iterating
                # without v are already at the top of the stack, with correct
                # values

                # We also build the subset with v and so we add values at the
                # top of the stack
                level += 3
                bitset_copy(stack.rows[level], current)
                bitset_add(stack.rows[level], v)
                bitset_copy(stack.rows[level + 1], left)
                bitset_union(stack.rows[level + 2], boundary, DG.rows[v])

                # We yield that new subset
                vertices = [int_to_vertex[a] for a in range(u, n)
                            if bitset_in(stack.rows[level], a)]
                if not exactly_k or bitset_len(current) == mk - 1:
                    if vertices_only:
                        yield vertices
                    else:
                        H = G.subgraph(vertices)
                        if induced:
                            if edges_only:
                                yield H.edges(sort=False, labels=labels)
                            else:
                                yield H
                        else:
                            # We use a decomposition into biconnected components to
                            # work on smaller graphs.
                            if H.is_directed():
                                blocks = H.to_undirected().blocks_and_cut_vertices()[0]
                            else:
                                blocks = H.blocks_and_cut_vertices()[0]
                            if len(blocks) == 1:
                                # H is strongly connected or biconnected
                                yield from connected_full_subgraphs(H, edges_only=edges_only,
                                                                    labels=labels)
                            else:
                                L = []
                                for bloc in blocks:
                                    if len(bloc) == 2:
                                        bb = [[e] for e in H.edge_boundary(bloc, bloc, labels=labels)]
                                        if len(bb) == 2:
                                            # H is directed with edges (u, v) and (v, u)
                                            bb.append(H.edge_boundary(bloc, bloc, labels=labels))
                                        L.append(bb)
                                    else:
                                        L.append(connected_full_subgraphs(H.subgraph(vertices=bloc),
                                                                          edges_only=True, labels=labels))

                                for edges in product(*L):
                                    good_edges = flatten(edges, ltypes=list)
                                    if edges_only:
                                        yield list(good_edges)
                                    else:
                                        yield H.subgraph(vertices=H, edges=good_edges)

            else:
                # We cannot extend the current subset, either due to a lack of
                # candidate (v == -1), or because the current subset has maximum
                # size. We pop
                level -= 3

    sig_on()
    binary_matrix_free(stack)
    binary_matrix_free(DG)
    sig_off()



# sage_setup: distribution = sagemath-graphs
r"""
Free monoid generated by prime knots available via the
:class:`~sage.knots.knotinfo.KnotInfoBase` class.

A generator of this free abelian monoid is a prime knot according to
the list at `KnotInfo <https://knotinfo.math.indiana.edu/>`__. A fully
amphicheiral prime knot is represented by exactly one generator with
the corresponding name. For non-chiral prime knots, there are
additionally one or three generators with the suffixes ``m``, ``r``
and ``c`` which specify the mirror and reverse images according to
their symmetry type.

AUTHORS:

- Sebastian Oehms June 2024: initial version
"""

##############################################################################
#       Copyright (C) 2024 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
##############################################################################

from sage.knots.knotinfo import SymmetryMutant
from sage.monoids.indexed_free_monoid import (IndexedFreeAbelianMonoid,
                                              IndexedFreeAbelianMonoidElement)
from sage.misc.cachefunc import cached_method
from sage.structure.unique_representation import UniqueRepresentation


class FreeKnotInfoMonoidElement(IndexedFreeAbelianMonoidElement):
    """
    An element of an indexed free abelian monoid.
    """
    def as_knot(self):
        r"""
        Return the knot represented by ``self``.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: FKIM.inject_variables(select=3)
            Defining K3_1
            Defining K3_1m
            sage: K = K3_1^2 * K3_1m
            sage: K.as_knot()
            Knot represented by 9 crossings
        """
        wl = self.to_word_list()
        P = self.parent()
        if len(wl) == 1:
            name = wl[0]
            L = P._index_dict[name][0].link()
            if name.endswith(SymmetryMutant.mirror_image.value):
                return L.mirror_image()
            if name.endswith(SymmetryMutant.reverse.value):
                return L.reverse()
            if name.endswith(SymmetryMutant.concordance_inverse.value):
                return L.mirror_image().reverse()
            return L
        else:
            from sage.misc.misc_c import prod
            return prod(P.gen(wl[i]).as_knot() for i in range(len(wl)))

    def to_knotinfo(self):
        r"""
        Return a word representing ``self`` as a list of pairs.

        Each pair ``(ki, sym)`` consists of a
        :class:`~sage.knots.knotinfo.KnotInfoBase` instance ``ki`` and
        :class:`~sage.knots.knotinfo.SymmetryMutant` instance ``sym``.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: FKIM.inject_variables(select=3)
            Defining K3_1
            Defining K3_1m
            sage: K = K3_1^2 * K3_1m
            sage: K.to_knotinfo()
            [(<KnotInfo.K3_1: '3_1'>, <SymmetryMutant.itself: 's'>),
            (<KnotInfo.K3_1: '3_1'>, <SymmetryMutant.itself: 's'>),
            (<KnotInfo.K3_1: '3_1'>, <SymmetryMutant.mirror_image: 'm'>)]
        """
        wl = self.to_word_list()
        P = self.parent()
        return [P._index_dict[w] for w in wl]


class FreeKnotInfoMonoid(IndexedFreeAbelianMonoid):

    Element = FreeKnotInfoMonoidElement

    @staticmethod
    def __classcall_private__(cls, max_crossing_number=6, prefix=None, **kwds):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FreeKnotInfoMonoid()
            Free abelian monoid of knots with at most 6 crossings
            sage: FreeKnotInfoMonoid(5)
            Free abelian monoid of knots with at most 5 crossings
        """
        if not prefix:
            prefix = 'KnotInfo'
        # We skip the IndexedMonoid__classcall__
        return UniqueRepresentation.__classcall__(cls, max_crossing_number,
                                                  prefix=prefix, **kwds)

    def __init__(self, max_crossing_number, category=None, prefix=None, **kwds):
        r"""
        Initialize ``self`` with generators belonging to prime knots with
        at most ``max_crossing_number`` crossings.

        TESTS:

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: FKIM4 =  FreeKnotInfoMonoid(4)
            sage: TestSuite(FKIM).run()
            sage: TestSuite(FKIM4).run()
        """
        self._max_crossing_number = None
        self._set_index_dictionary(max_crossing_number=max_crossing_number)
        from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
        indices = FiniteEnumeratedSet(self._index_dict)
        super().__init__(indices, prefix)

    def _from_knotinfo(self, knotinfo, symmetry_mutant):
        r"""
        Return the name on the generator for the given ``symmetry_mutant``
        of the given entry ``knotinfo`` if the KnotInfo database.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: from sage.knots.knotinfo import SymmetryMutant
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: ki = KnotInfo.K5_2
            sage: FKIM._from_knotinfo(ki, SymmetryMutant.itself)
            'K5_2'
            sage: FKIM._from_knotinfo(ki, SymmetryMutant.concordance_inverse)
            'K5_2c'
        """
        if symmetry_mutant == SymmetryMutant.itself:
            return knotinfo.name
        else:
            return '%s%s' % (knotinfo.name, symmetry_mutant.value)

    def _set_index_dictionary(self, max_crossing_number=6):
        r"""
        Set or expand the set of generators.
        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FreeKnotInfoMonoid()
            Free abelian monoid of knots with at most 6 crossings

        TESTS::

            sage: from sage.features.databases import DatabaseKnotInfo
            sage: F = DatabaseKnotInfo()
            sage: F.hide()
            sage: FreeKnotInfoMonoid(7)  # indirect doctest
            Traceback (most recent call last):
            ...
            sage.features.FeatureNotPresentError: database_knotinfo is not available.
            Feature `database_knotinfo` is hidden.
            Use method `unhide` to make it available again.
            sage: F.unhide()
        """
        if max_crossing_number > 6:
            from sage.features.databases import DatabaseKnotInfo
            DatabaseKnotInfo().require()

        current_max_crossing_number = self._max_crossing_number
        if not current_max_crossing_number:
            current_max_crossing_number = - 1
            self._index_dict = {}
        self._max_crossing_number = max_crossing_number

        def add_index(ki, sym):
            self._index_dict[self._from_knotinfo(ki, sym)] = (ki, sym)

        from sage.knots.knotinfo import KnotInfo
        for K in KnotInfo:
            ncr = K.crossing_number()
            if ncr <= current_max_crossing_number:
                continue
            if ncr > self._max_crossing_number:
                break
            for sym in SymmetryMutant:
                if sym.is_minimal(K):
                    add_index(K, sym)
        if current_max_crossing_number > 0:
            from sage.sets.finite_enumerated_set import FiniteEnumeratedSet
            self._indices = FiniteEnumeratedSet(self._index_dict)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

          sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
          sage: FreeKnotInfoMonoid(4)
          Free abelian monoid of knots with at most 4 crossings
        """
        return "Free abelian monoid of knots with at most %s crossings" % self._max_crossing_number

    def _element_constructor_(self, x=None):
        """
        Create an element of this abelian monoid from ``x``.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: K = KnotInfo.K5_1.link().mirror_image()
            sage: FKIM(K)                                                               # needs libhomfly sage.modules
            KnotInfo['K5_1m']
        """
        if isinstance(x, tuple):
            if len(x) == 2:
                ki, sym = x
                from sage.knots.knotinfo import KnotInfoBase
                if isinstance(ki, KnotInfoBase) and isinstance(sym, SymmetryMutant):
                    mcr = ki.crossing_number()
                    if mcr > self._max_crossing_number:
                        self._set_index_dictionary(max_crossing_number=mcr)

                    sym_min = min([sym] + sym.matches(ki))
                    return self.gen(self._from_knotinfo(ki, sym_min))

        from sage.knots.knot import Knot
        from sage.knots.link import Link
        if not isinstance(x, Knot):
            if isinstance(x, Link):
                x = Knot(x.pd_code())
        if isinstance(x, Knot):
            return self.from_knot(x)
        return self.element_class(self, x)

    @cached_method
    def _check_elements(self, knot, elems):
        r"""
        Return a matching item from the list in ``elems`` if it exists.
        Elsewise return ``None``. This is a helper method for .meth:`from_knot`.

        INPUT:

        - ``knot`` -- an instance of :class:`~sage.knots.knot.Knot`
        - ``elems`` -- a tuple of elements of ``self``

        EXAMPLES::

            sage: # needs sage.groups
            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: FKIM.inject_variables(select=3)
            Defining K3_1
            Defining K3_1m
            sage: elems = (K3_1, K3_1m)
            sage: K = Knots().from_table(3, 1)
            sage: FKIM._check_elements(K, elems)
            KnotInfo['K3_1m']
            sage: K = Knots().from_table(4, 1)
            sage: FKIM._check_elements(K, elems) is None
            True
        """
        for e in elems:
            k = e.as_knot()
            if knot.pd_code() == k.pd_code():
                return e
            if knot._markov_move_cmp(k.braid()):
                return e
        return None

    @cached_method
    def _search_composition(self, max_cr, knot, hpoly):
        r"""
        Add KnotInfo items to the list of candidates that have
        matching Homfly polynomial.

        INPUT:

        -  ``max_cr`` -- max number of crossing to stop searching
        -  ``knot`` -- instance of :class:`~sage.knots.knot.Knot`
        -  ``hpoly`` -- Homfly polynomial to search for a component

        OUTPUT:

        A tuple of elements of ``self`` that match a (not necessarily prime or
        proper) component of the given knot having the given Homfly polynomial.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: FKIM.inject_variables(select=3)
            Defining K3_1
            Defining K3_1m
            sage: KI = K3_1 * K3_1m
            sage: K = KI.as_knot()
            sage: h = K3_1.to_knotinfo()[0][0].homfly_polynomial()                      # needs libhomfly sage.modules
            sage: FKIM._search_composition(3, K, h)                                     # needs libhomfly sage.modules
            (KnotInfo['K3_1'],)
        """
        from sage.knots.knotinfo import KnotInfo

        def hp_mirr(hp):
            v, z = hp.parent().gens()
            return hp.subs({v: ~v, z: z})

        former_cr = 3
        res = []
        for K in KnotInfo:
            if not K.is_knot():
                break
            c = K.crossing_number()
            if c < 3:
                continue
            if c > max_cr:
                break
            hp = K.homfly_polynomial()
            hp_sym = {s: hp for s in SymmetryMutant if s.is_minimal(K)}
            hpm = hp_mirr(hp)
            if hp != hpm:
                hp_sym[SymmetryMutant.mirror_image] = hpm
                if SymmetryMutant.concordance_inverse in hp_sym.keys():
                    hp_sym[SymmetryMutant.concordance_inverse] = hpm

            for sym_mut, hps in hp_sym.items():
                if hps.divides(hpoly):
                    Kgen = self((K, sym_mut))
                    h = hpoly // hps
                    if h.is_unit():
                        res += [Kgen]
                    else:
                        res_rec = self._search_composition(max_cr - c, knot, h)
                        if res_rec:
                            res += [Kgen * k for k in res_rec]
            if c > former_cr and res:
                k = self._check_elements(knot, tuple(res))
                if k:
                    # matching item found
                    return tuple([k])
                former_cr = c

        return tuple(sorted(set(res)))

    @cached_method
    def _from_knot(self, knot):
        """
        Create a tuple of element of this abelian monoid which possibly
        represent ``knot``. This method caches the performance relevant
        part of :meth:`from_knot`.

        INPUT:

        - ``knot`` -- an instance of :class:`~sage.knots.knot.Knot`

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: K = KnotInfo.K5_1.link().mirror_image()
            sage: FKIM._from_knot(K)                                                    # needs sage.groups sage.modules
            (KnotInfo['K5_1m'],)
        """
        hp = knot.homfly_polynomial(normalization='vz')
        return self._search_composition(13, knot, hp)

    def from_knot(self, knot, unique=True):
        """
        Create an element of this abelian monoid from ``knot``.

        INPUT:

        - ``knot`` -- an instance of :class:`~sage.knots.knot.Knot`

        - ``unique`` -- boolean (default is ``True``). This only affects the case
          where a unique identification is not possible. If set to ``False`` you
          can obtain a matching list (see explanation of the output below)

        OUTPUT:

        An instance of the element class of ``self`` per default. If the keyword
        argument ``unique`` then a list of such instances is returned.

        EXAMPLES::

            sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
            sage: FKIM =  FreeKnotInfoMonoid()
            sage: K = KnotInfo.K5_1.link().mirror_image()
            sage: FKIM.from_knot(K)                                                     # needs libhomfly sage.modules
            KnotInfo['K5_1m']

            sage: # optional - database_knotinfo
            sage: K = Knot(KnotInfo.K9_12.braid())
            sage: FKIM.from_knot(K)                   # long time                       # optional - database_knotinfo
            Traceback (most recent call last):
            ...
            NotImplementedError: this (possibly non prime) knot cannot be
            identified uniquely by KnotInfo
            use keyword argument `unique` to obtain more details
            sage: FKIM.from_knot(K, unique=False)     # long time                       # optional - database_knotinfo
            [KnotInfo['K4_1']*KnotInfo['K5_2'], KnotInfo['K9_12']]
        """
        hp = knot.homfly_polynomial(normalization='vz')
        num_summands = sum(e for _, e in hp.factor())
        if num_summands == 1:
            return knot.get_knotinfo()

        res = self._from_knot(knot)
        if res:
            if len(res) == 1:
                if unique:
                    return res[0]
                return [res[0]]  # to be consistent with get_knotinfo
            k = self._check_elements(knot, res)
            if k:
                if unique:
                    return k
                return [k]  # to be consistent with get_knotinfo

        if res and not unique:
            return sorted(set(res))
        if unique and len(res) > 1:
            non_unique_hint = '\nuse keyword argument `unique` to obtain more details'
            raise NotImplementedError('this (possibly non prime) knot cannot be identified uniquely by KnotInfo%s' % non_unique_hint)
        raise NotImplementedError('this (possibly non prime) knot cannot be identified by KnotInfo')

    def inject_variables(self, select=None, verbose=True):
        """
        Inject ``self`` with its name into the namespace of the
        Python code from which this function is called.

        INPUT:

        - ``select`` -- instance of :class:`~sage.knots.knotinfo.KnotInfoBase`,
          :class:`~sage.knots.knotinfo.KnotInfoSeries` or an integer. In all
          cases the input is used to restrict the injected generators to the
          according subset (number of crossings in the case of integer)
        - ``verbose`` -- boolean (optional, default ``True``) to suppress
          the message printed on the invocation

        EXAMPLES::

          sage: from sage.knots.free_knotinfo_monoid import FreeKnotInfoMonoid
          sage: FKIM = FreeKnotInfoMonoid(5)
          sage: FKIM.inject_variables(select=3)
          Defining K3_1
          Defining K3_1m
          sage: FKIM.inject_variables(select=KnotInfo.K5_2)
          Defining K5_2
          Defining K5_2m
          sage: FKIM.inject_variables(select=KnotInfo.K5_2.series())
          Defining K5_1
          Defining K5_1m
          sage: FKIM.inject_variables()
          Defining K0_1
          Defining K4_1
        """
        from sage.knots.knotinfo import KnotInfoBase, KnotInfoSeries
        from sage.rings.integer import Integer
        gen_list = []
        idx_dict = self._index_dict
        max_crn = self._max_crossing_number
        gens = self.gens()
        if select:
            if isinstance(select, KnotInfoBase):
                crn = select.crossing_number()
                if crn > max_crn:
                    self._set_index_dictionary(max_crossing_number=crn)
                gen_list += [k for k, v in idx_dict.items() if v[0] == select]
            elif isinstance(select, KnotInfoSeries):
                for v in select:
                    self.inject_variables(select=v)
                return
            elif type(select) is int or isinstance(select, Integer):
                crn = select
                if crn > max_crn:
                    self._set_index_dictionary(max_crossing_number=crn)
                gen_list += [k for k, v in idx_dict.items()
                             if v[0].crossing_number() == crn]
            else:
                raise TypeError('cannot select generators by %s' % select)
        else:
            gen_list = list(idx_dict.keys())

        from sage.repl.user_globals import set_global, get_globals
        for name in gen_list:
            if name not in get_globals().keys():
                set_global(name, gens[name])
                if verbose:
                    print("Defining %s" % (name))



# sage_setup: distribution = sagemath-graphs
r"""
Functions for reading/building graphs/digraphs

This module gathers functions needed to build a graph from any other data.

.. NOTE::

    This is an **internal** module of Sage. All features implemented here are
    made available to end-users through the constructors of :class:`Graph` and
    :class:`DiGraph`.

Note that because they are called by the constructors of :class:`Graph` and
:class:`DiGraph`, most of these functions modify a graph inplace.

{INDEX_OF_FUNCTIONS}

Functions
---------
"""
from sage.cpython.string import bytes_to_str
from sage.misc.rest_index_of_methods import gen_rest_table_index
import sys


def from_graph6(G, g6_string):
    r"""
    Fill ``G`` with the data of a graph6 string.

    INPUT:

    - ``G`` -- a graph

    - ``g6_string`` -- a graph6 string

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_graph6
        sage: g = Graph()
        sage: from_graph6(g, 'IheA@GUAo')
        sage: g.is_isomorphic(graphs.PetersenGraph())
        True
    """
    from .generic_graph_pyx import length_and_string_from_graph6, binary_string_from_graph6

    if isinstance(g6_string, bytes):
        g6_string = bytes_to_str(g6_string)
    elif not isinstance(g6_string, str):
        raise ValueError('if input format is graph6, then g6_string must be a string')
    n = g6_string.find('\n')
    if n == -1:
        n = len(g6_string)
    ss = g6_string[:n]
    n, s = length_and_string_from_graph6(ss)
    m = binary_string_from_graph6(s, n)
    expected = n * (n - 1) // 2 + (6 - n * (n - 1) // 2) % 6
    if len(m) > expected:
        raise RuntimeError("the string (%s) seems corrupt: for n = %d, the string is too long" % (ss, n))
    elif len(m) < expected:
        raise RuntimeError("the string (%s) seems corrupt: for n = %d, the string is too short" % (ss, n))
    G.add_vertices(range(n))
    k = 0
    for i in range(n):
        for j in range(i):
            if m[k] == '1':
                G._backend.add_edge(i, j, None, False)
            k += 1


def from_sparse6(G, g6_string):
    r"""
    Fill ``G`` with the data of a sparse6 string.

    INPUT:

    - ``G`` -- a graph

    - ``g6_string`` -- a sparse6 string

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_sparse6
        sage: g = Graph()
        sage: from_sparse6(g, ':I`ES@obGkqegW~')
        sage: g.is_isomorphic(graphs.PetersenGraph())
        True
    """
    from .generic_graph_pyx import length_and_string_from_graph6, int_to_binary_string

    if isinstance(g6_string, bytes):
        g6_string = bytes_to_str(g6_string)
    elif not isinstance(g6_string, str):
        raise ValueError('if input format is graph6, then g6_string must be a string')

    n = g6_string.find('\n')
    if n == -1:
        n = len(g6_string)
    s = g6_string[:n]
    n, s = length_and_string_from_graph6(s[1:])
    if not n:
        edges = []
    else:
        from sage.rings.integer_ring import ZZ
        k = int((ZZ(n) - 1).nbits())
        ords = [ord(i) for i in s]
        if any(o > 126 or o < 63 for o in ords):
            raise RuntimeError("the string seems corrupt: valid characters are \n" + ''.join(chr(i) for i in range(63, 127)))
        bits = ''.join(int_to_binary_string(o-63).zfill(6) for o in ords)
        if not k:
            b = [int(x) for x in bits]
            x = [0] * len(b)
        else:
            b = []
            x = []
            for i in range(0, len(bits)-k, k+1):
                b.append(int(bits[i:i+1], 2))
                x.append(int(bits[i+1:i+k+1], 2))
        v = 0
        edges = []
        for i in range(len(b)):
            v += b[i]  # +1 if b[i] == 1 else 0
            if x[i] > v:
                v = x[i]
            else:
                if v < n:
                    edges.append((x[i], v))
    G.add_vertices(range(n))
    G.add_edges(edges)


def from_dig6(G, dig6_string):
    r"""
    Fill ``G`` with the data of a dig6 string.

    INPUT:

    - ``G`` -- a graph

    - ``dig6_string`` -- a dig6 string

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_dig6
        sage: g = DiGraph()
        sage: from_dig6(g, digraphs.Circuit(10).dig6_string())
        sage: g.is_isomorphic(digraphs.Circuit(10))
        True

    The string may represent a directed graph with loops::

        sage: L = DiGraph(loops=True)
        sage: from_dig6(L, 'CW`C')
        sage: L.edges(labels=False, sort=True)
        [(0, 1), (0, 2), (1, 2), (2, 3), (3, 3)]
    """
    from .generic_graph_pyx import length_and_string_from_graph6, binary_string_from_dig6
    if isinstance(dig6_string, bytes):
        dig6_string = bytes_to_str(dig6_string)
    elif not isinstance(dig6_string, str):
        raise ValueError('if input format is dig6, then dig6_string must be a string')
    n = dig6_string.find('\n')
    if n == -1:
        n = len(dig6_string)
    ss = dig6_string[:n]
    n, s = length_and_string_from_graph6(ss)
    m = binary_string_from_dig6(s, n)
    expected = n**2
    if len(m) > expected:
        raise RuntimeError("the string (%s) seems corrupt: for n = %d, the string is too long" % (ss, n))
    elif len(m) < expected:
        raise RuntimeError("the string (%s) seems corrupt: for n = %d, the string is too short" % (ss, n))
    G.add_vertices(range(n))
    k = 0
    for i in range(n):
        for j in range(n):
            if m[k] == '1':
                G._backend.add_edge(i, j, None, True)
            k += 1


def from_seidel_adjacency_matrix(G, M):
    r"""
    Fill ``G`` with the data of a Seidel adjacency matrix.

    INPUT:

    - ``G`` -- a graph

    - ``M`` -- a Seidel adjacency matrix

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_seidel_adjacency_matrix
        sage: g = Graph()
        sage: sam = graphs.PetersenGraph().seidel_adjacency_matrix()                    # needs sage.modules
        sage: from_seidel_adjacency_matrix(g, sam)                                      # needs sage.modules
        sage: g.is_isomorphic(graphs.PetersenGraph())                                   # needs sage.modules
        True
    """
    from sage.structure.element import Matrix
    from sage.rings.integer_ring import ZZ
    assert isinstance(M, Matrix)

    if M.base_ring() != ZZ:
        try:
            M = M.change_ring(ZZ)
        except TypeError:
            raise ValueError("the adjacency matrix of a Seidel graph must" +
                             " have only 0,1,-1 integer entries")

    if M.is_sparse():
        entries = set(M[i, j] for i, j in M.nonzero_positions())
    else:
        entries = set(M.list())

    if any(e < -1 or e > 1 for e in entries):
        raise ValueError("the adjacency matrix of a Seidel graph must" +
                         " have only 0,1,-1 integer entries")
    if any(i == j for i, j in M.nonzero_positions()):
        raise ValueError("the adjacency matrix of a Seidel graph must" +
                         " have 0s on the main diagonal")
    if not M.is_symmetric():
        raise ValueError("the adjacency matrix of a Seidel graph must be symmetric")

    G.add_vertices(range(M.nrows()))
    G.add_edges((i, j) for i, j in M.nonzero_positions() if i <= j and M[i, j] < 0)


def from_adjacency_matrix(G, M, loops=False, multiedges=False, weighted=False):
    r"""
    Fill ``G`` with the data of an adjacency matrix.

    INPUT:

    - ``G`` -- a :class:`Graph` or :class:`DiGraph`

    - ``M`` -- an adjacency matrix

    - ``loops``, ``multiedges``, ``weighted`` -- booleans (default: ``False``);
      whether to consider the graph as having loops, multiple edges, or weights

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_adjacency_matrix
        sage: g = Graph()
        sage: from_adjacency_matrix(g, graphs.PetersenGraph().adjacency_matrix())       # needs sage.modules
        sage: g.is_isomorphic(graphs.PetersenGraph())                                   # needs sage.modules
        True
    """
    from sage.structure.element import Matrix
    from sage.rings.integer_ring import ZZ
    assert isinstance(M, Matrix)
    # note: the adjacency matrix might be weighted and hence not
    # necessarily consists of integers
    if not weighted and M.base_ring() != ZZ:
        try:
            M = M.change_ring(ZZ)
        except TypeError:
            if weighted is False:
                raise ValueError("the adjacency matrix of a non-weighted graph" +
                                 " must have only nonnegative integer entries")
            weighted = True

    if M.is_sparse():
        entries = set(M[i, j] for i, j in M.nonzero_positions())
    else:
        entries = set(M.list())

    if not weighted and any(e < 0 for e in entries):
        if weighted is False:
            raise ValueError("the adjacency matrix of a non-weighted graph" +
                             " must have only nonnegative integer entries")
        weighted = True
        if multiedges is None:
            multiedges = False
    if weighted is None:
        weighted = False

    if multiedges is None:
        multiedges = ((not weighted) and any(e != 0 and e != 1 for e in entries))

    if not loops and any(M[i, i] for i in range(M.nrows())):
        if loops is False:
            raise ValueError("the adjacency matrix of a non-weighted graph" +
                             " must have zeroes on the diagonal")
        loops = True
    if loops is None:
        loops = False
    G.allow_loops(loops, check=False)
    G.allow_multiple_edges(multiedges, check=False)
    G.add_vertices(range(M.nrows()))
    if G.is_directed():
        pairs = M.nonzero_positions()
    else:
        pairs = ((i, j) for i, j in M.nonzero_positions() if i <= j)
    if weighted:
        G.add_edges((i, j, M[i][j]) for i, j in pairs)
    elif multiedges:
        G.add_edges((i, j) for i, j in pairs for _ in range(int(M[i][j])))
    else:
        G.add_edges((i, j) for i, j in pairs)
    G._weighted = weighted


def from_incidence_matrix(G, M, loops=False, multiedges=False, weighted=False):
    r"""
    Fill ``G`` with the data of an incidence matrix.

    INPUT:

    - ``G`` -- a graph

    - ``M`` -- an incidence matrix

    - ``loops``, ``multiedges``, ``weighted`` -- booleans (default: ``False``);
      whether to consider the graph as having loops, multiple edges, or weights

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_incidence_matrix
        sage: g = Graph()
        sage: from_incidence_matrix(g, graphs.PetersenGraph().incidence_matrix())       # needs sage.modules
        sage: g.is_isomorphic(graphs.PetersenGraph())                                   # needs sage.modules
        True
    """
    from sage.structure.element import Matrix
    assert isinstance(M, Matrix)

    oriented = any(M[pos] < 0 for pos in M.nonzero_positions(copy=False))

    positions = []
    for i in range(M.ncols()):
        NZ = M.nonzero_positions_in_column(i)
        if len(NZ) == 1:
            if oriented:
                raise ValueError("column {} of the (oriented) incidence "
                                 "matrix contains only one nonzero value".format(i))
            elif M[NZ[0], i] != 2:
                raise ValueError("each column of a non-oriented incidence "
                                 "matrix must sum to 2, but column {} does not".format(i))
            if loops is None:
                loops = True
            positions.append((NZ[0], NZ[0]))
        elif (len(NZ) != 2 or
              (oriented and not ((M[NZ[0], i] == +1 and M[NZ[1], i] == -1) or
                                 (M[NZ[0], i] == -1 and M[NZ[1], i] == +1))) or
              (not oriented and (M[NZ[0], i] != 1 or M[NZ[1], i] != 1))):
            msg = "there must be one or two nonzero entries per column in an incidence matrix, "
            msg += "got entries {} in column {}".format([M[j, i] for j in NZ], i)
            raise ValueError(msg)
        else:
            positions.append(tuple(NZ))

    if weighted is None:
        G._weighted = False
    if multiedges is None:
        total = len(positions)
        multiedges = len(set(positions)) < total
    G.allow_loops(False if loops is None else loops, check=False)
    G.allow_multiple_edges(multiedges, check=False)
    G.add_vertices(range(M.nrows()))
    G.add_edges(positions)


def from_oriented_incidence_matrix(G, M, loops=False, multiedges=False, weighted=False):
    r"""
    Fill ``G`` with the data of an *oriented* incidence matrix.

    An oriented incidence matrix is the incidence matrix of a directed graph, in
    which each non-loop edge corresponds to a `+1` and a `-1`, indicating its
    source and destination.

    INPUT:

    - ``G`` -- a :class:`DiGraph`

    - ``M`` -- an incidence matrix

    - ``loops``, ``multiedges``, ``weighted`` -- booleans (default: ``False``);
      whether to consider the graph as having loops, multiple edges, or weights

    .. NOTE:: ``weighted`` is currently ignored.

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_oriented_incidence_matrix
        sage: g = DiGraph()
        sage: im = digraphs.Circuit(10).incidence_matrix()                              # needs sage.modules
        sage: from_oriented_incidence_matrix(g, im)                                     # needs sage.modules
        sage: g.is_isomorphic(digraphs.Circuit(10))                                     # needs sage.modules
        True

    TESTS:

    Fix bug reported in :issue:`22985`::

        sage: DiGraph(matrix ([[1,0,0,1],[0,0,1,1],[0,0,1,1]]).transpose())             # needs sage.modules
        Traceback (most recent call last):
        ...
        ValueError: each column represents an edge: -1 goes to 1

    Handle incidence matrix containing a column with only zeros (:issue:`29275`)::

        sage: m = Matrix([[0,1],[0,-1],[0,0]]); m                                       # needs sage.modules
        [ 0  1]
        [ 0 -1]
        [ 0  0]
        sage: G = DiGraph(m, format='incidence_matrix')                                 # needs sage.modules
        sage: list(G.edges(sort=True, labels=False))                                    # needs sage.modules
        [(1, 0)]

    Handle incidence matrix [[1],[-1]] (:issue:`29275`)::

        sage: m = Matrix([[1],[-1]]); m                                                 # needs sage.modules
        [ 1]
        [-1]
        sage: G = DiGraph(m, format='incidence_matrix')                                 # needs sage.modules
        sage: list(G.edges(sort=True, labels=False))                                    # needs sage.modules
        [(1, 0)]
    """
    from sage.structure.element import Matrix
    assert isinstance(M, Matrix)

    positions = []
    for c in M.columns():
        NZ = c.nonzero_positions()
        if not NZ:
            continue
        if len(NZ) != 2:
            raise ValueError("there must be two nonzero entries (-1 & 1) per column")
        L = sorted([c[i] for i in NZ])
        if L != [-1, 1]:
            raise ValueError("each column represents an edge: -1 goes to 1")
        if c[NZ[0]] == -1:
            positions.append(tuple(NZ))
        else:
            positions.append((NZ[1], NZ[0]))
    if multiedges is None:
        total = len(positions)
        multiedges = len(set(positions)) < total
    G.allow_loops(bool(loops), check=False)
    G.allow_multiple_edges(multiedges, check=False)
    G.add_vertices(range(M.nrows()))
    G.add_edges(positions)


def from_dict_of_dicts(G, M, loops=False, multiedges=False, weighted=False, convert_empty_dict_labels_to_None=False):
    r"""
    Fill ``G`` with the data of a dictionary of dictionaries.

    INPUT:

    - ``G`` -- a graph

    - ``M`` -- dictionary of dictionaries

    - ``loops``, ``multiedges``, ``weighted`` -- booleans (default: ``False``);
      whether to consider the graph as having loops, multiple edges, or weights

    - ``convert_empty_dict_labels_to_None`` -- booleans (default: ``False``);
      whether to adjust for empty dicts instead of ``None`` in NetworkX default
      edge labels

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_dict_of_dicts
        sage: g = Graph()
        sage: from_dict_of_dicts(g, graphs.PetersenGraph().to_dictionary(edge_labels=True))
        sage: g.is_isomorphic(graphs.PetersenGraph())
        True

    The resulting order of vertices is unspecified but deterministic::

        sage: from sage.graphs.graph_input import from_dict_of_dicts
        sage: g = Graph()
        sage: from_dict_of_dicts(g, {i: {} for i in range(99, 90, -1)})
        sage: g.vertices(sort=False)
        [99, 98, 97, 96, 95, 94, 93, 92, 91]

    TESTS:

    :issue:`32831` is fixed::

        sage: DiGraph({0: {}, 1: {}, 2: {}, 3: {}, 4: {}})
        Digraph on 5 vertices
    """
    if any(not isinstance(M[u], dict) for u in M):
        raise ValueError("input dict must be a consistent format")

    if not loops:
        if any(u in neighb for u, neighb in M.items()):
            if loops is False:
                u = next(u for u, neighb in M.items() if u in neighb)
                raise ValueError("the graph was built with loops=False but input M has a loop at {}".format(u))
            loops = True
        if loops is None:
            loops = False
    if weighted is None:
        G._weighted = False
    input_multiedges = multiedges
    if multiedges is not False:
        if not all(isinstance(M[u][v], list) for u in M for v in M[u]):
            if multiedges:
                raise ValueError("dict of dicts for multigraph must be in the format {v: {u: list}}")
            multiedges = False
        if multiedges is None and M:
            multiedges = True

    G.allow_loops(loops, check=False)
    G.allow_multiple_edges(multiedges, check=False)
    # Use keys of a dictionary instead of a set, to preserve insertion order
    verts = dict(M)
    for d in M.values():
        verts.update(d)
    G.add_vertices(verts.keys())
    if convert_empty_dict_labels_to_None:
        def relabel(x):
            return x if x != {} else None
    else:
        def relabel(x):
            return x

    is_directed = G.is_directed()
    if not is_directed and multiedges:
        v_to_id = {v: i for i, v in enumerate(verts.keys())}
        for u in M:
            for v in M[u]:
                if v_to_id[u] <= v_to_id[v] or v not in M or u not in M[v] or u == v:
                    for label in M[u][v]:
                        G._backend.add_edge(u, v, relabel(label), False)
    elif multiedges:
        for u in M:
            for v in M[u]:
                for label in M[u][v]:
                    G._backend.add_edge(u, v, relabel(label), is_directed)
    else:
        for u in M:
            for v in M[u]:
                G._backend.add_edge(u, v, relabel(M[u][v]), is_directed)
    if not G.size() and input_multiedges is not True:
        G.allow_multiple_edges(False, check=False)


def from_dict_of_lists(G, D, loops=False, multiedges=False, weighted=False):
    r"""
    Fill ``G`` with the data of a dictionary of lists.

    INPUT:

    - ``G`` -- a :class:`Graph` or :class:`DiGraph`

    - ``D`` -- dictionary of lists

    - ``loops``, ``multiedges``, ``weighted`` -- booleans (default: ``False``);
      whether to consider the graph as having loops, multiple edges, or weights

    EXAMPLES::

        sage: from sage.graphs.graph_input import from_dict_of_lists
        sage: g = Graph()
        sage: from_dict_of_lists(g, graphs.PetersenGraph().to_dictionary())
        sage: g.is_isomorphic(graphs.PetersenGraph())
        True

    The resulting order of vertices is unspecified but deterministic::

        sage: from sage.graphs.graph_input import from_dict_of_lists
        sage: g = Graph()
        sage: from_dict_of_lists(g, {i: [] for i in range(99, 90, -1)})
        sage: g.vertices(sort=False)
        [99, 98, 97, 96, 95, 94, 93, 92, 91]
    """
    # Use keys of a dictionary instead of a set, to preserve insertion order
    verts = dict(D)
    verts.update({v: None for l in D.values() for v in l})
    if not loops:
        if any(u in neighb for u, neighb in D.items()):
            if loops is False:
                u = next(u for u, neighb in D.items() if u in neighb)
                raise ValueError("the graph was built with loops=False but input D has a loop at {}".format(u))
            loops = True
        if loops is None:
            loops = False
    if weighted is None:
        G._weighted = False
    if not multiedges:
        for u in D:
            if len(set(D[u])) != len(D[u]):
                if multiedges is False:
                    v = next(v for v in D[u] if D[u].count(v) > 1)
                    raise ValueError("non-multigraph got several edges (%s, %s)" % (u, v))
                multiedges = True
                break
        if multiedges is None:
            multiedges = False
    G.allow_loops(loops, check=False)
    G.allow_multiple_edges(multiedges, check=False)
    G.add_vertices(verts.keys())

    is_directed = G.is_directed()
    if not is_directed and multiedges:
        v_to_id = {v: i for i, v in enumerate(verts.keys())}
        for u in D:
            for v in D[u]:
                if (v_to_id[u] <= v_to_id[v] or
                        v not in D or u not in D[v] or u == v):
                    G._backend.add_edge(u, v, None, False)
    else:
        for u in D:
            for v in D[u]:
                G._backend.add_edge(u, v, None, is_directed)


def from_networkx_graph(G, gnx, weighted=None, loops=None, multiedges=None,
                        convert_empty_dict_labels_to_None=None):
    r"""
    Fill `G` with the data of a NetworkX (di)graph.

    INPUT:

    - ``G`` -- a :class:`Graph` or :class:`DiGraph`

    - ``gnx`` -- a NetworkX ``Graph``, ``MultiGraph``, ``DiGraph`` or
      ``MultiDiGraph``

    - ``weighted`` -- boolean (default: ``None``); whether graph thinks of
      itself as weighted or not. See
      :meth:`~sage.graphs.generic_graph.GenericGraph.weighted`.

    - ``loops`` -- boolean (default: ``None``); whether to allow loops

    - ``multiedges`` -- boolean (default: ``None``); whether to allow multiple
      edges

    - ``convert_empty_dict_labels_to_None`` -- boolean (default: ``None``);
      whether to replace the default edge labels used by NetworkX (empty
      dictionaries) by ``None``, the default Sage edge label. When set to
      ``False``, empty dictionaries are not converted to ``None``.

    EXAMPLES:

    Feeding a :class:`Graph` with a NetworkX ``Graph``::

        sage: # needs networkx
        sage: from sage.graphs.graph_input import from_networkx_graph
        sage: import networkx
        sage: G = Graph()
        sage: _ = gnx = networkx.Graph()
        sage: _ = gnx.add_edge(0, 1)
        sage: _ = gnx.add_edge(1, 2)
        sage: from_networkx_graph(G, gnx)
        sage: G.edges(sort=True, labels=False)
        [(0, 1), (1, 2)]

    Feeding a :class:`Graph` with a NetworkX ``MultiGraph``::

        sage: # needs networkx
        sage: G = Graph()
        sage: gnx = networkx.MultiGraph()
        sage: _ = gnx.add_edge(0, 1)
        sage: _ = gnx.add_edge(0, 1)
        sage: from_networkx_graph(G, gnx)
        sage: G.edges(sort=True, labels=False)
        [(0, 1), (0, 1)]
        sage: G = Graph()
        sage: from_networkx_graph(G, gnx, multiedges=False)
        sage: G.edges(sort=True, labels=False)
        [(0, 1)]

    When feeding a :class:`Graph` `G` with a NetworkX ``DiGraph`` `D`, `G` has
    one edge `(u, v)` whenever `D` has arc `(u, v)` or `(v, u)` or both::

        sage: # needs networkx
        sage: G = Graph()
        sage: D = networkx.DiGraph()
        sage: _ = D.add_edge(0, 1)
        sage: from_networkx_graph(G, D)
        sage: G.edges(sort=True, labels=False)
        [(0, 1)]
        sage: G = Graph()
        sage: _ = D.add_edge(1, 0)
        sage: from_networkx_graph(G, D)
        sage: G.edges(sort=True, labels=False)
        [(0, 1)]

    When feeding a :class:`Graph` `G` with a NetworkX ``MultiDiGraph`` `D`, the
    number of edges between `u` and `v` in `G` is the maximum between the number
    of arcs `(u, v)` and the number of arcs `(v, u)` in D`::

        sage: # needs networkx
        sage: G = Graph()
        sage: D = networkx.MultiDiGraph()
        sage: _ = D.add_edge(0, 1)
        sage: _ = D.add_edge(1, 0)
        sage: _ = D.add_edge(1, 0)
        sage: D.edges()
        OutMultiEdgeDataView([(0, 1), (1, 0), (1, 0)])
        sage: from_networkx_graph(G, D)
        sage: G.edges(sort=True, labels=False)
        [(0, 1), (0, 1)]

    Feeding a :class:`DiGraph` with a NetworkX ``DiGraph``::

        sage: # needs networkx
        sage: from sage.graphs.graph_input import from_networkx_graph
        sage: import networkx
        sage: G = DiGraph()
        sage: _ = gnx = networkx.DiGraph()
        sage: _ = gnx.add_edge(0, 1)
        sage: _ = gnx.add_edge(1, 2)
        sage: from_networkx_graph(G, gnx)
        sage: G.edges(sort=True, labels=False)
        [(0, 1), (1, 2)]

    Feeding a :class:`DiGraph` with a NetworkX ``MultiDiGraph``::

        sage: # needs networkx
        sage: G = DiGraph()
        sage: gnx = networkx.MultiDiGraph()
        sage: _ = gnx.add_edge(0, 1)
        sage: _ = gnx.add_edge(0, 1)
        sage: from_networkx_graph(G, gnx)
        sage: G.edges(sort=True, labels=False)
        [(0, 1), (0, 1)]
        sage: G = DiGraph()
        sage: from_networkx_graph(G, gnx, multiedges=False)
        sage: G.edges(sort=True, labels=False)
        [(0, 1)]

    When feeding a :class:`DiGraph` `G` with a NetworkX ``Graph`` `H`, `G` has
    both arcs `(u, v)` and `(v, u)` if `G` has edge `(u, v)`::

        sage: # needs networkx
        sage: G = DiGraph()
        sage: H = networkx.Graph()
        sage: _ = H.add_edge(0, 1)
        sage: from_networkx_graph(G, H)
        sage: G.edges(labels=False, sort=True)
        [(0, 1), (1, 0)]

    When feeding a :class:`DiGraph` `G` with a NetworkX ``MultiGraph`` `H`, `G`
    has `k` arcs `(u, v)` and `k` arcs `(v, u)` if `H` has `k` edges `(u, v)`,
    unless parameter ``multiedges`` is set to ``False``::

        sage: # needs networkx
        sage: G = DiGraph()
        sage: H = networkx.MultiGraph()
        sage: _ = H.add_edge(0, 1)
        sage: _ = H.add_edge(0, 1)
        sage: _ = H.add_edge(0, 1)
        sage: H.edges()
        MultiEdgeDataView([(0, 1), (0, 1), (0, 1)])
        sage: from_networkx_graph(G, H)
        sage: G.edges(labels=False, sort=True)
        [(0, 1), (0, 1), (0, 1), (1, 0), (1, 0), (1, 0)]
        sage: G = DiGraph()
        sage: from_networkx_graph(G, H, multiedges=False)
        sage: G.edges(labels=False, sort=True)
        [(0, 1), (1, 0)]

    TESTS:

    The first parameter must be a :class:`Graph` or :class:`DiGraph`::

        sage: from sage.graphs.graph_input import from_networkx_graph
        sage: from_networkx_graph("foo", "bar")
        Traceback (most recent call last):
        ...
        ValueError: the first parameter must a Sage Graph or DiGraph

    The second parameter must be a NetworkX ``Graph``, ``MultiGraph``,
      ``DiGraph`` or ``MultiDiGraph``::

        sage: from sage.graphs.graph_input import from_networkx_graph
        sage: from_networkx_graph(Graph(), "bar")                                       # needs networkx
        Traceback (most recent call last):
        ...
        ValueError: the second parameter must be a NetworkX (Multi)(Di)Graph
    """
    from sage.graphs.graph import Graph
    from sage.graphs.digraph import DiGraph
    if not isinstance(G, (Graph, DiGraph)):
        raise ValueError("the first parameter must a Sage Graph or DiGraph")
    import networkx
    if not isinstance(gnx, (networkx.Graph, networkx.DiGraph)):
        raise ValueError("the second parameter must be a NetworkX (Multi)(Di)Graph")

    if G.is_directed() != gnx.is_directed():
        if gnx.is_directed():
            gnx = gnx.to_undirected()
        else:
            gnx = gnx.to_directed()

    if weighted is None:
        if multiedges is None:
            multiedges = gnx.is_multigraph()
        if loops is None:
            loops = any(u == v for u, v in gnx.edges())

    G.allow_loops(loops, check=False)
    G.allow_multiple_edges(multiedges, check=False)
    G.add_vertices(gnx.nodes())
    G.set_vertices(gnx.nodes(data=True))
    if convert_empty_dict_labels_to_None is not False:
        def r(label):
            return None if label == {} else label
        G.add_edges((u, v, r(ll)) for u, v, ll in gnx.edges(data=True))
    else:
        G.add_edges(gnx.edges(data=True))


__doc__ = __doc__.format(INDEX_OF_FUNCTIONS=gen_rest_table_index(sys.modules[__name__]))
